#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


import json
import os
import pathlib

import orjson
from typing import Dict, List, Set, Tuple, Union

from PySide6.QtCore import (QLineF, QPoint, QPointF, QRect, QRectF,
                            QRegularExpression, Qt, Signal, Slot)
from PySide6.QtGui import (QFont, QFontDatabase, QPen, QTextDocument)
from PySide6.QtWidgets import (QComboBox, QDialog, QGraphicsItem,
                               QGraphicsRectItem, QGraphicsScene,
                               QGraphicsSceneMouseEvent)

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.editFunctions as edf
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.backend.undoStack as us
import revedaEditor.checks.schematic as schk
import revedaEditor.common.labels as lbl
import revedaEditor.common.net as snet
import revedaEditor.common.shapes as shp  # import the shapes
import revedaEditor.fileio.loadJSON as lj
import revedaEditor.fileio.schematicEncoder as schenc
import revedaEditor.gui.alignItems as alg
import revedaEditor.gui.fileDialogues as fd
import revedaEditor.gui.propertyDialogues as pdlg
from revedaEditor.backend.pdkLoader import importPDKModule
from revedaEditor.scenes.editorScene import editorScene

schlyr = importPDKModule('schLayers')


class schematicScene(editorScene):
    wireEditFinished = Signal(snet.schematicNet)
    alignLineFinished = Signal(shp.alignLine)
    stretchNet = Signal(snet.schematicNet, str)
    SCHEMATIC_SHAPES = (snet.schematicNet, shp.schematicPin, shp.text,
                        shp.schematicSymbol)

    def __init__(self, parent):
        super().__init__(parent)

        # Initialize counters
        self.instCounter = 0
        self.instanceCounter = 0
        self.netCounter = 0

        # Initialize modes with default values
        self.editModes = ddef.schematicModes(selectItem=True, deleteItem=False,
                                             moveItem=False, copyItem=False,
                                             rotateItem=False, changeOrigin=False,
                                             panView=False, zoomView=False, drawPin=False,
                                             drawWire=False, drawBus=False,
                                             drawText=False, addInstance=False,
                                             stretchItem=False, nameNet=False,
                                             alignItems=False)
        self.messages = {'selectItem': 'Select Item', 'deleteItem': 'Delete Item',
                         'moveItem': 'Move Item', 'copyItem': 'Copy Item',
                         'rotateItem': 'Rotate Item',
                         'changeOrigin': 'Change Origin',
                         'panView': 'Pan View at mouse Press Position',
                         'zoomView': 'Draw rectangle to zoom in',
                         'stretchItem': 'Stretch Item', 'drawPin': 'Draw Pin',
                         'drawWire': 'Draw Wire', 'drawBus': 'Draw Bus',
                         'drawText': 'Draw Text', 'addInstance': 'Add Instance',
                         'nameNet': 'Name Net', 'drawSymbol': 'Draw Symbol',
                         'alignItems': 'Select alignment option.'}

        self.selectModes = ddef.schematicSelectModes(selectAll=True,
                                                     selectDevice=False,
                                                     selectNet=False,
                                                     selectPin=False)

        # Initialize selection trackers
        self.selectedNet = None
        self.selectedPin = None
        self.selectedSymbol = None
        self.selectedSymbolPin = None

        # Initialize pin defaults
        self.pinName = ""
        self.pinType = "Signal"
        self.pinDir = "Input"

        # Initialize internal state trackers
        self._newNet = None
        self._stretchNet = None
        self._newInstance = None
        self._newPin = None
        self._newText = None
        self._newNetNameObj = None
        self.textTuple = None
        self.netNameString = None
        self.newInstanceTuple = None
        self.newAlignLine = None
        self._snapPointRect = self.defineSnapRect()

        # error shapes
        self.overlapRectSet = set()

        # Initialize view properties
        self.highlightNets = False
        self.hierarchyTrail = ""
        self.parentEditor = None
        self.parentObj = None

        # Font initialization
        self._initializeFont()

        # Connect signals
        self.wireEditFinished.connect(self._handleWireFinished)
        self.stretchNet.connect(self._handleStretchNet)
        self.alignLineFinished.connect(alg.alignToLine)

        # Initialize cache
        self._symbolCache = {}

    def _initializeFont(self):
        """Initialize fixed-width font settings."""
        fontFamilies = QFontDatabase.families(QFontDatabase.WritingSystem.Latin)

        # Find first fixed-pitch font family
        fixedFamily = next(family for family in fontFamilies if
                           QFontDatabase.isFixedPitch(family))

        # Get font style and create font
        fontStyle = QFontDatabase.styles(fixedFamily)[1]
        self.fixedFont = QFont(fixedFamily)
        self.fixedFont.setStyleName(fontStyle)

        # Set font size
        fontSize = QFontDatabase.pointSizes(fixedFamily, fontStyle)[3]
        self.fixedFont.setPointSize(fontSize)
        self.fixedFont.setKerning(False)

    def defineSnapRect(self):
        snapPointRect = QGraphicsRectItem()
        snapPointRect.setRect(QRect(-2, -2, 4, 4))
        snapPointRect.setZValue(100)
        snapPointRect.setVisible(False)
        if schlyr and hasattr(schlyr, 'guideLinePen'):
            snapPointRect.setPen(schlyr.guideLinePen)

        return snapPointRect

    @property
    def drawMode(self):
        return any((self.editModes.drawPin, self.editModes.drawWire,
                    self.editModes.drawText))

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if self.selectedItemGroup and self.editModes.moveItem:
            for item in self.selectedItemGroup.childItems():
                if isinstance(item, shp.schematicSymbol) or isinstance(item,
                                                                       shp.schematicPin):
                    item.generatePinNetDict()
                    item.initializeSnapLines(self)
                elif isinstance(item, snet.schematicNet):
                    item.generateEndPointNetDict()
                    item.initializeSnapLines()

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """Handle mouse move event."""
        super().mouseMoveEvent(event)
        self.mouseMoveLoc = event.scenePos().toPoint()

        if self._newInstance and self.editModes.addInstance:
            self._newInstance.setPos(self.mouseMoveLoc)
        elif self._newPin and self.editModes.drawPin:

            self._newPin.setPos(self.mouseMoveLoc - self._newPin.start)
        elif self._newNet and (self.editModes.drawWire or self.editModes.drawBus):
            netEndPoint = self.findSnapPoint(self.mouseMoveLoc, {self._newNet})
            self._snapPointRect.setPos(netEndPoint)
            self._newNet.draftLine = QLineF(self._newNet.draftLine.p1(),
                                            netEndPoint)
        elif self.newAlignLine and self.editModes.alignItems:
            self.newAlignLine.draftLine = QLineF(
                self.newAlignLine.draftLine.p1(), self.mouseMoveLoc)
        elif self._stretchNet and self.editModes.stretchItem:
            netEndPoint = self.findSnapPoint(self.mouseMoveLoc, set())
            self._snapPointRect.setVisible(True)
            self._snapPointRect.setPos(netEndPoint)
            self._stretchNet.draftLine = QLineF(self._stretchNet.draftLine.p1(),
                                                netEndPoint)
        elif self._newText and self.editModes.drawText:

            self._newText.start = self.mouseMoveLoc
        elif self.editModes.nameNet and self._newNetNameObj:
            self._newNetNameObj.setPos(self.mouseMoveLoc)
        elif self.editModes.moveItem and self.selectedItemGroup:
            for item in self.selectedItemGroup.childItems():
                if isinstance(item, (shp.schematicSymbol, shp.schematicPin,
                                     snet.schematicNet)):
                    item.updateSnapLines()

        cursorPosition = self.snapToGrid(self.mouseMoveLoc - self.origin)
        self.statusLine.showMessage(
            f"Cursor Position: ({cursorPosition.x()}, {cursorPosition.y()})")
        self.messageLine.setText(self.messages[self.editModes.mode()])

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent):
        # Finalize snap lines for group net moves before the group is destroyed in super()
        if self.editModes.moveItem and self.selectedItemGroup:
            for item in self.selectedItemGroup.childItems():
                if isinstance(item, snet.schematicNet):
                    item._finishSnapLines()
        return super().mouseReleaseEvent(event)

    def _handleMouseRelease(self, mousePos: QPoint,
                            button: Qt.MouseButton) -> None:
        """
        Handle mouse release logic.

        :param mouseReleaseLoc: QPoint instance
        :param button: Qt.MouseButton instance
        """

        if button == Qt.MouseButton.LeftButton:
            # modifiers = QGuiApplication.keyboardModifiers()
            if self.editModes.addInstance:
                self._handleAddInstance(mousePos)
            elif self.editModes.drawPin:
                self._handleDrawPin(mousePos)
            elif self.editModes.drawWire:
                self._handleDrawWire(mousePos)
            elif self.editModes.drawBus:
                self._handleDrawBus(mousePos)
            elif self.editModes.drawText:
                self._handleDrawText(mousePos)
            elif self.editModes.rotateItem:
                self.rotateSelectedItems(mousePos)
            elif self.editModes.stretchItem:
                self._handleStretchItem()
            elif self.editModes.nameNet:
                self._handleNameNet(mousePos)
            elif self.editModes.alignItems:
                self._handleAlignItemLine(mousePos)

    def _handleAddInstance(self, eventLoc: QPoint) -> None:
        """
        Handle add instance logic.

        :param eventLoc: QPoint instance
        """
        if self._newInstance:
            self._newInstance = None
        self._newInstance = self.drawInstance(self.newInstanceTuple, eventLoc)
        self._newInstance.setSelected(True)

    def _handleDrawPin(self, mouseReleaseLoc: QPoint) -> None:
        """
        Handle draw pin logic.

        :param mouseReleaseLoc: QPoint instance
        """
        # if self._newPin:
        #     self._newPin = None
        self._newPin = self.addPin(mouseReleaseLoc)
        self._newPin.setSelected(True)

    def _handleDrawWire(self, eventLoc: QPoint) -> None:
        """Handle draw wire logic with continuous mode."""
        ignoredSet = {self._newNet} if self._newNet is not None else set()
        snapPoint = self.snapToGrid(self.findSnapPoint(eventLoc, ignoredSet))

        if self._newNet is None:
            # Start new net
            self._snapPointRect.setPos(snapPoint)
            self._newNet = snet.schematicNet(snapPoint, snapPoint, 0)
            self.addUndoStack(self._newNet)
        else:
            # Continue from last point – use same snapPoint for segment end and next start
            self._newNet.draftLine = QLineF(self._newNet.draftLine.p1(), snapPoint)
            self.wireEditFinished.emit(self._newNet)

            # Start next segment from same endpoint
            self._snapPointRect.setPos(snapPoint)
            self._newNet = snet.schematicNet(snapPoint, snapPoint, 0)
            self.addUndoStack(self._newNet)

    def _handleAlignItemLine(self, eventLoc: QPoint) -> None:
        if self.newAlignLine is None:
            from PySide6.QtWidgets import (QApplication, )
            alignDlg = [w for w in QApplication.topLevelWidgets() if isinstance(w,
                alg.alignItemsDialogue) and w.isVisible() and w.scene == self][
                0]
            if alignDlg.horizontalAlignButton.isChecked():
                self.newAlignLine = shp.alignLine(QLineF(eventLoc, eventLoc), 1,
                                                  0)  # horizontal
            else:
                self.newAlignLine = shp.alignLine(QLineF(eventLoc, eventLoc), 1,
                                                  1)
            self.addUndoStack(self.newAlignLine)
        else:
            self.newAlignLine.draftLine = QLineF(
                self.newAlignLine.draftLine.p1(), eventLoc)
            self.alignLineFinished.emit(self.newAlignLine)
            # self.newAlignLine = None
            self.editModes.setMode('selectItem')

    def _handleDrawBus(self, eventLoc: QPoint):
        if self._newNet:
            self._newNet.draftLine = QLineF(self._newNet.draftLine.p1(),
                                            self.snapToGrid(eventLoc))
            self.wireEditFinished.emit(self._newNet)
            self._newNet = None
        startSnapPoint = self.snapToGrid(self.findSnapPoint(eventLoc, set()))
        self._newNet = snet.schematicNet(startSnapPoint, startSnapPoint, 1)
        self.addUndoStack(self._newNet)

    def _handleDrawText(self, mouseReleaseLoc: QPoint) -> None:
        """
        Handle draw text logic.

        :param mouseReleaseLoc: QPoint instance
        """
        if self._newText:
            self._newText = None
            self.textTuple = None
        if self.textTuple:
            self._newText = self.addNote(mouseReleaseLoc, self.textTuple)

    def _handleStretchItem(self):
        if self._stretchNet:
            self._stretchNet.stretch = False
            self._stretchNet = None
            self._snapPointRect.setVisible(False)
            self.editModes.stretchItem = False

    def _handleNameNet(self, mouseReleaseLoc):
        if self.netNameString:
            self._newNetNameObj = snet.netName(self.netNameString)
            self._newNetNameObj.setPos(mouseReleaseLoc)
            self.addUndoStack(self._newNetNameObj)
            self.netNameString = None
        elif self._newNetNameObj and self.selectedNet:
            self.selectedNet.name = self._newNetNameObj.name
            self.undoStack.undo()
            # self.undoStack.removeLastCommand()
            self.selectedNet.nameStrength = snet.netNameStrengthEnum.SET
            self.selectedNet.setSelected(False)
            self.selectedNet = None

    def updateStretchNet(self):
        self._stretchNet.draftLine = QLineF(self._stretchNet.draftLine.p1(),
                                            self.mouseMoveLoc)

    @Slot(snet.schematicNet)
    def _handleWireFinished(self, newNet: snet.schematicNet):
        """
        check if the new net is valid. If it has zero length, remove it. Otherwise process it.

        """

        if newNet.draftLine.length() < self.snapGrid / 2:
            self.removeItem(newNet)
            self.undoStack.removeLastCommand()
        else:
            newNetSceneRect = newNet.sceneBoundingRect().adjusted(-self.snapGrid,
                                                                  -self.snapGrid,
                                                                  self.snapGrid,
                                                                  self.snapGrid)
            self.mergeSplitNets(newNet)
            self.invalidate(newNetSceneRect, QGraphicsScene.BackgroundLayer)

    def _updateNets(self, nets_to_remove: set, nets_to_add: set, name_source_net):
        """Helper to update nets in scene"""
        for net in nets_to_remove:
            self.removeItem(net)
        for net in nets_to_add:
            self.addItem(net)
            net.mergeNetName(name_source_net)

    def mergeSplitNets(self, inputNet: snet.schematicNet):
        merged, outputNet, processedNets = self.mergeNets(inputNet)

        if merged:
            splitDone, splitOutputNets = self.splitInputNet(outputNet)
            if splitDone:
                self._updateNets(processedNets - splitOutputNets,
                                 splitOutputNets - processedNets, outputNet)
            else:
                for net in processedNets:
                    self.removeItem(net)
                self.addItem(outputNet)
        else:
            splitDone, splitOutputNets = self.splitInputNet(inputNet)
            if splitDone:
                self._updateNets({inputNet} - splitOutputNets,
                                 splitOutputNets - {inputNet}, inputNet)

    def mergeNets(self, inputNet: snet.schematicNet) -> Tuple[
        bool, snet.schematicNet, Set[snet.schematicNet]]:
        """
        Merges overlapping nets and returns the merged net.
        """
        otherNets = inputNet.findOverlapNets()
        if not otherNets:
            return False, inputNet, set()

        parallelNets = {net for net in otherNets if inputNet.isParallel(net)}
        if not parallelNets:
            return False, inputNet, set()

        points = list(inputNet.sceneEndPoints)  # Create a copy
        busExists = int(any(net.width for net in parallelNets) or inputNet.width)

        for net in parallelNets:
            points.extend(net.sceneEndPoints)

        furthestPoints = self.findFurthestPoints(points)
        mergedNet = snet.schematicNet(*furthestPoints, width=busExists)

        processedNets = parallelNets | {inputNet}
        for net in processedNets:
            mergedNet.mergeNetName(net)

        return True, mergedNet, processedNets

    def splitInputNet(self, inputNet) -> tuple[bool, Set[snet.schematicNet]]:
        sceneShapeRect = inputNet.sceneShapeRect
        sceneItems = self.items(sceneShapeRect)

        splitPointsSet = set()

        # Single iteration to collect all relevant points
        for item in sceneItems:
            if isinstance(item, snet.schematicNet) and inputNet.isOrthogonal(
                    item):
                for point in item.sceneEndPoints:
                    if sceneShapeRect.contains(point):
                        splitPointsSet.add(point)
            elif isinstance(item, (shp.symbolPin, shp.schematicPin)):
                splitPointsSet.add(item.mapToScene(item.start).toPoint())

        if not splitPointsSet:
            return False, set()

        # Create ordered points list
        splitPointsList = [inputNet.sceneEndPoints[0], *splitPointsSet,
                           inputNet.sceneEndPoints[1], ]
        orderedPoints = list(dict.fromkeys(self.orderPoints(splitPointsList)))

        # Create split nets
        splitNetSet = set()
        is_selected = inputNet.isSelected()

        for i in range(len(orderedPoints) - 1):
            splitNet = snet.schematicNet(orderedPoints[i], orderedPoints[i + 1],
                                         inputNet.width)
            if not splitNet.draftLine.isNull():
                if is_selected:
                    splitNet.setSelected(True)
                splitNetSet.add(splitNet)

        return True, splitNetSet

    def findSnapPoint(self, eventLoc: QPoint,
                      ignoredSet: set[snet.schematicNet]) -> QPoint:
        snapRect = QRect(eventLoc.x() - self.snapTuple[0],
                         eventLoc.y() - self.snapTuple[1], 4 * self.snapTuple[0],
                         4 * self.snapTuple[1], )
        snapPoints = self.findConnectPoints(snapRect, ignoredSet)

        if self._newNet:
            snapPoints.update(self.findNetInterSect(self._newNet, snapRect))
        if snapPoints:
            lengths = [(snapPoint - eventLoc).manhattanLength() for snapPoint in
                       snapPoints]
            closestPoint = list(snapPoints)[lengths.index(min(lengths))]
            return closestPoint
        else:
            return eventLoc

    def findConnectPoints(self, sceneRect: QRect,
                          ignoredSet: set[QGraphicsItem]) -> set[QPoint]:
        snapPoints = set()
        rectItems = set(self.items(sceneRect)) - ignoredSet
        for item in rectItems:
            if isinstance(item, snet.schematicNet) and any(
                    list(map(sceneRect.contains, item.sceneEndPoints))):
                snapPoints.add(item.sceneEndPoints[list(
                    map(sceneRect.contains, item.sceneEndPoints)).index(True)])
            elif isinstance(item, shp.symbolPin):
                snapPoints.add(item.mapToScene(item.start).toPoint())
            elif isinstance(item, shp.schematicPin):
                snapPoints.add(item.mapToScene(item.start).toPoint())
        return snapPoints

    def findNetInterSect(self, inputNet: snet.schematicNet, rect: QRect) -> set[
        QPoint]:
        # Find all nets in the rectangle except the input net
        netsInSnapRectSet = {netItem for netItem in self.items(rect) if
                             isinstance(netItem,
                                        snet.schematicNet) and netItem.isOrthogonal(
                                 inputNet)}
        snapPointsSet = set()
        l1 = QLineF(inputNet.sceneEndPoints[0], inputNet.sceneEndPoints[1])
        unitVector = l1.unitVector()
        dx = unitVector.dx() if unitVector.dx() else 0
        dy = unitVector.dy() if unitVector.dy() else 0
        newEndX = l1.x2() + rect.width() * dx
        newEndY = l1.y2() + rect.height() * dy
        extendedL1 = QLineF(l1.p1(), QPointF(newEndX, newEndY))
        for netItem in netsInSnapRectSet:
            l2 = QLineF(netItem.sceneEndPoints[0], netItem.sceneEndPoints[1])
            (_, intersectPoint) = extendedL1.intersects(l2)
            if intersectPoint:
                snapPointsSet.add(intersectPoint.toPoint())
        return snapPointsSet

    # def findNetStretchPoints(self, netItem: snet.schematicNet,
    #                          snapDistance: int) -> dict[int, QPoint]:
    #     netEndPointsDict: dict[int, QPoint] = {}
    #     sceneEndPoints = netItem.sceneEndPoints
    #     for netEnd in sceneEndPoints:
    #         snapRect: QRect = QRect(netEnd.x() - snapDistance,
    #                                 netEnd.y() - snapDistance, 2 * snapDistance,
    #                                 2 * snapDistance, )
    #         snapRectItems = set(self.items(snapRect)) - {netItem}
    #
    #         for item in snapRectItems:
    #             if isinstance(item, snet.schematicNet) and any(
    #                     list(map(snapRect.contains, item.sceneEndPoints))):
    #                 netEndPointsDict[sceneEndPoints.index(netEnd)] = netEnd
    #             elif (isinstance(item,
    #                              shp.symbolPin | shp.schematicPin)) and snapRect.contains(
    #                 item.mapToScene(item.start).toPoint()):
    #                 netEndPointsDict[
    #                     sceneEndPoints.index(netEnd)] = item.mapToScene(
    #                     item.start).toPoint()
    #             if netEndPointsDict.get(sceneEndPoints.index(
    #                     netEnd)):  # after finding one point, no need to iterate.
    #                 break
    #     return netEndPointsDict

    @staticmethod
    def orderPoints(points: list[QPoint]) -> list[QPoint]:
        currentPoint = points.pop(0)
        orderedPoints = [currentPoint]

        while points:
            distances = [(point - currentPoint).manhattanLength() for point in
                         points]
            nearest_point_index = distances.index(min(distances))
            nearestPoint = points[nearest_point_index]
            orderedPoints.append(nearestPoint)
            currentPoint = points.pop(nearest_point_index)

        return orderedPoints

    @staticmethod
    def findFurthestPoints(points: list[QPoint]) -> tuple[QPoint, QPoint]:
        """
        Find the two points with the maximum distance between them.
        """
        max_distance = 0
        furthest_points = None

        for i in range(len(points)):
            for j in range(i + 1, len(points)):
                distance = (points[i] - points[j]).manhattanLength()
                if distance > max_distance:
                    max_distance = distance
                    furthest_points = (points[i], points[j])

        return furthest_points[0], furthest_points[1]

    def _handleStretchNet(self, netItem: snet.schematicNet, stretchEnd: str):
        match stretchEnd:
            case "p2":
                self._stretchNet = snet.schematicNet(netItem.sceneEndPoints[0],
                                                     netItem.sceneEndPoints[1])
            case "p1":
                self._stretchNet = snet.schematicNet(netItem.sceneEndPoints[1],
                                                     netItem.sceneEndPoints[0])
        self._stretchNet.stretch = True
        self._stretchNet.mergeNetName(netItem)
        addDeleteStretchNetCommand = us.addDeleteShapeUndo(self, self._stretchNet,
                                                           netItem)
        self.undoStack.push(addDeleteStretchNetCommand)

    def generatePinNetMap(self, sceneSymbolSet: set[shp.schematicSymbol]):
        """
        For symbols in sceneSymbolSet, find which pin is connected to which net.
        Handles both single pins and bus connections.
        """
        for symbolItem in sceneSymbolSet:
            self.genSymbolPinNetMap(symbolItem)

    def findSceneSymbolSet(self) -> set[shp.schematicSymbol]:
        """
        Find all the symbols on the scene as a set.
        """
        return {item for item in self.items() if
                isinstance(item, shp.schematicSymbol)}

    def findSceneNetsSet(self) -> set[snet.schematicNet]:
        return {item for item in self.items() if
                isinstance(item, snet.schematicNet)}

    def findRectSymbolPin(self, rect: Union[QRect, QRectF]) -> set[shp.symbolPin]:
        pinsRectSet = {item for item in self.items(rect) if
                       isinstance(item, shp.symbolPin)}
        return pinsRectSet

    def findRectSchemPins(self, rect: Union[QRect, QRectF]) -> set[
        shp.schematicPin]:
        pinsRectSet = {item for item in self.items(rect) if
                       isinstance(item, shp.schematicPin)}
        return pinsRectSet

    def findSceneSchemPinsSet(self) -> set[shp.schematicPin]:
        pinsSceneSet = {item for item in self.items() if
                        isinstance(item, shp.schematicPin)}
        if pinsSceneSet:  # check pinsSceneSet is empty
            return pinsSceneSet
        else:
            return set()

    def addStretchWires(self, start: QPoint, end: QPoint) -> List[
        "snet.schematicNet"]:
        """
        Add a trio of wires between two points.

        Args:
            start (QPoint): The starting point of the wire.
            end (QPoint): The ending point of the wire.

        Returns:
            List[snet.schematicNet]: A list of schematic net objects representing the wires.
        """
        try:
            if start == end:
                self.logger.warning(
                    "Start and end points are the same. No wire added.")
                return []

            if start.y() == end.y() or start.x() == end.x():
                # Horizontal or vertical line
                return [snet.schematicNet(start, end)]

            # Calculate intermediate points
            firstPointX = self.snapToBase((end.x() - start.x()) / 3 + start.x(),
                                          self.snapTuple[0])
            firstPoint = QPoint(firstPointX, start.y())
            secondPoint = QPoint(firstPointX, end.y())

            # Create wire segments
            lines = []
            segments = [(start, firstPoint), (firstPoint, secondPoint),
                        (secondPoint, end), ]
            for seg_start, seg_end in segments:
                if seg_start != seg_end:
                    lines.append(snet.schematicNet(seg_start, seg_end))

            return lines

        except Exception as e:
            self.logger.error(f"Error in addStretchWires: {e}", exc_info=True)
            return []

    def addPin(self, pos: QPoint) -> shp.schematicPin:
        try:
            pin = shp.schematicPin(pos, self.pinName, self.pinDir, self.pinType)
            self.addUndoStack(pin)
            return pin
        except Exception as e:
            self.logger.error(f"Pin add error: {e}")

    def addNote(self, pos: QPoint, textTuple: Tuple) -> shp.text:
        """
        Changed the method name not to clash with qgraphicsscene addText method.
        """
        text = shp.text(pos, *textTuple)
        self.addUndoStack(text)
        return text

    def drawInstance(self, instanceTuple: ddef.viewNameTuple, pos: QPoint):
        """
        Add an instance of a symbol to the scene.
        """
        instance = self.instSymbol(instanceTuple, pos)
        if instance:  # Add check for None
            self.addUndoStack(instance)
            self.instanceCounter += 1
            return instance
        else:
            return None

    def instSymbol(self, instanceTuple: ddef.viewNameTuple, pos: QPoint):
        viewItem = libm.findViewItem(self.editorWindow.libraryView.libraryModel,
                                     instanceTuple.libraryName,
                                     instanceTuple.cellName,
                                     instanceTuple.viewName, )
        viewPath = viewItem.viewPath
        try:
            # Try to get items from cache first
            items = self._symbolCache.get(viewPath)
            if items is None:
                with open(viewPath, "r") as temp:
                    items = json.load(temp)
                    self._symbolCache[viewPath] = items

            # Use comprehensions for better performance
            itemAttributes = {item["nam"]: item["def"] for item in items[2:] if
                              item["type"] == "attr"}

            itemShapes = [lj.symbolItems(self).create(item) for item in items[2:]
                          if item["type"] != "attr"]
            symbolInstance = shp.schematicSymbol(itemShapes, itemAttributes)
            cellItem = viewItem.parent()
            libItem = cellItem.parent()
            # Batch property assignments
            instanceProperties = {"pos": pos, "counter": self.instanceCounter,
                                  "instanceName": f"I{self.instanceCounter}",
                                  "libraryName": libItem.libraryName,
                                  "cellName": cellItem.cellName,
                                  "viewName": viewItem.viewName, }

            for prop, value in instanceProperties.items():
                setattr(symbolInstance, prop, value)
            # as pylabels can depend on NLPLabel results, process first
            # NLP Labels.
            for labelItem in symbolInstance.labels.values():
                if labelItem.labelType == lbl.symbolLabel.labelTypes[1]:
                    labelItem.labelDefs()
            for labelItem in symbolInstance.labels.values():
                if labelItem.labelType == lbl.symbolLabel.labelTypes[2]:
                    labelItem.labelDefs()

            return symbolInstance

        except FileNotFoundError:
            self.logger.error(f"Symbol file not found: {viewPath}")
            return None
        except json.JSONDecodeError:
            self.logger.error(f"Invalid JSON in symbol file: {viewPath}")
            return None
        except Exception as e:
            self.logger.warning(f"instantiation error: {e}")
            return None

    def copySelectedItems(self):
        selectedItems = [item for item in self.selectedItems() if
                         item.parentItem() is None]
        if not selectedItems:
            return
        copyShapesList = []
        if selectedItems:
            for item in selectedItems:
                selectedItemJson = json.dumps(item, cls=schenc.schematicEncoder)
                itemCopyDict = json.loads(selectedItemJson)
                shape = lj.schematicItems(self).create(itemCopyDict)
                if shape is not None:
                    if isinstance(shape, shp.schematicSymbol):
                        self.instanceCounter += 1
                        shape.instanceName = f"I{self.instanceCounter}"
                        shape.counter = int(self.instanceCounter)
                        [label.labelDefs() for label in shape.labels.values()]
                    copyShapesList.append(shape)
            self.addListUndoStack(copyShapesList)
            self.selectedItemGroup = self.createItemGroup(copyShapesList)
            self.selectedItemGroup.setSelected(True)

    def saveSchematic(self, file: pathlib.Path) -> bool:
        """
        Save the schematic to a file with optimized memory usage and error handling.

        Args:
            file (pathlib.Path): The file path to save the schematic to.

        Raises:
            IOError: If there are file operation errors
            JSONEncodeError: If there are JSON serialization errors
        """
        try:
            self.itemsRefSet = set(self.items())
            # Ensure parentW directory exists
            file.parent.mkdir(parents=True, exist_ok=True)

            # Create temporary file in the same directory
            tempFile = file.with_suffix(".tmp")
            # Write to temporary file first
            with self.measureDuration():
                with tempFile.open(mode="w", buffering=8192) as f:
                    # Start array
                    f.write("[\n")

                    header_items = [{"viewType": "schematic"}, {
                        "snapGrid": (self.majorGrid, self.snapGrid)}, ]
                    json.dump(header_items[0], f)
                    f.write(",\n")
                    json.dump(header_items[1], f)
                    topLevelItems = filter(
                        lambda item: isinstance(item, self.SCHEMATIC_SHAPES),
                        self.items())
                    # Stream items
                    for item in list(topLevelItems):
                        f.write(",\n")
                        try:
                            json.dump(item, f, cls=schenc.schematicEncoder)
                        except Exception as json_err:
                            self.logger.error(
                                f"Failed to serialize item: {str(json_err)}")
                            return

                    # Close array
                    f.write("\n]")

                    # Ensure all data is written to disk
                    f.flush()
                    os.fsync(f.fileno())

                # Atomic file replacement
                tempFile.replace(file)

                self.logger.info(
                    f"Saved schematic to {self.editorWindow.cellName}:"
                    f"{self.editorWindow.viewName}")
                self.undoStack.clear()
                return True
        except IOError as io_err:
            self.logger.error(f"IO Error while saving schematic: {str(io_err)}")
            return False
        except Exception as e:
            self.logger.error(
                f"Unexpected error while saving schematic: {str(e)}")
            # Clean up temporary file if it exists
            if tempFile.exists():
                tempFile.unlink()
            return False
        finally:
            if tempFile.exists():
                tempFile.unlink()

    @staticmethod
    def findEditorTypeString(editorWindow):
        """
        This function returns the type of the parentW editor as a string.
        The type of the parentW editor is determined by finding the last dot in the
        string representation of the type of the parentW editor and returning the
        string after the last dot. If there is no dot in the string representation
        of the type of the parentW editor, the entire string is returned.
        """
        index: int = str(type(editorWindow)).rfind(".")
        if index == -1:
            return str(type(editorWindow))
        else:
            return str(type(editorWindow))[index + 1: -2]

    def loadDesign(self, filePathObj: pathlib.Path) -> None:
        """
        Load schematic from a JSON file and initialize grid settings.

        Args:
            filePathObj (pathlib.Path): Path to the schematic JSON file

        Raises:
            JSONDecodeError: If the file contains invalid JSON
            FileNotFoundError: If the specified file doesn't exist
            KeyError: If required grid settings are missing
        """
        try:
            with filePathObj.open("rb") as file:
                decodedData = orjson.loads(file.read())
            with self.measureDuration():
                if len(decodedData) < 2:
                    viewDict = decodedData[0] if decodedData else {}
                    gridSettings = None
                    itemData = []
                else:
                    viewDict, gridSettings, *itemData = decodedData

                if viewDict.get("viewType") != "schematic":
                    self.logger.error("Not a schematic file!")
                    return

                if gridSettings and gridSettings.get("snapGrid"):
                    self.editorWindow.configureGridSettings(
                        decodedData[1].get("snapGrid",
                                           (self.majorGrid, self.snapGrid)))
                self.blockSignals(True)
                try:
                    self.createSchematicItems(itemData)
                finally:
                    self.blockSignals(False)
            self.itemsRefSet = set(self.items())

        except (orjson.JSONDecodeError, FileNotFoundError) as e:
            self.logger.error(f"File error while loading schematic: {e}")
            return
        except KeyError as e:
            self.logger.error(f"Invalid schematic format - missing key: {e}")
            return
        except Exception as e:
            self.logger.error(f"Unexpected error loading schematic: {e}")
            return

    def createSchematicItems(self, itemsList: List[Dict]):
        factory = lj.schematicItems(self)
        for itemDict in itemsList:
            itemShape = factory.create(itemDict)
            if (isinstance(itemShape,
                           shp.schematicSymbol) and itemShape.counter > self.instanceCounter):
                self.instanceCounter = itemShape.counter + 1

            if itemShape is not None:
                self.addItem(itemShape)

    def viewObjProperties(self):
        """
        Display the properties of the selected object.
        """
        try:
            selectedItems = [item for item in self.selectedItems() if
                             item.parentItem() is None]
            if selectedItems:
                for item in selectedItems:
                    item.prepareGeometryChange()
                    if isinstance(item, shp.schematicSymbol):
                        self.setInstanceProperties(item)
                    elif isinstance(item, snet.schematicNet):
                        self.setNetProperties(item)
                    elif isinstance(item, shp.text):
                        self.setTextProperties(item)
                    elif isinstance(item, shp.schematicPin):
                        self.setSchematicPinProperties(item)
                    elif isinstance(item, snet.netName):
                        self.setNetProperties(item.parentItem())
        except Exception as e:
            self.logger.error(e)

    def setInstanceProperties(self, item: shp.schematicSymbol):
        dlg = pdlg.instanceProperties(self.editorWindow)
        dlg.libNameEdit.setText(item.libraryName)
        dlg.cellNameEdit.setText(item.cellName)
        dlg.cellNameEdit.setEnabled(True)

        dlg.viewNameEdit.setText(item.viewName)
        dlg.instNameEdit.setText(item.instanceName)
        location = (item.scenePos() - self.origin).toTuple()
        dlg.xLocationEdit.setText(str(location[0]))
        dlg.yLocationEdit.setText(str(location[1]))
        dlg.angleEdit.setText(str(item.angle))
        row_index = 0
        # iterate through the item labels.
        for label in item.labels.values():
            if label.labelDefinition not in lbl.symbolLabel.predefinedLabels:
                dlg.instanceLabelsLayout.addWidget(
                    edf.boldLabel(label.labelName[1:], dlg), row_index, 0)
                labelValueEdit = edf.longLineEdit()
                labelValueEdit.setText(str(label.labelValue))
                dlg.instanceLabelsLayout.addWidget(labelValueEdit, row_index, 1)
                visibleCombo = QComboBox(dlg)
                visibleCombo.setInsertPolicy(QComboBox.NoInsert)
                visibleCombo.addItems(["True", "False"])
                if label.labelVisible:
                    visibleCombo.setCurrentIndex(0)
                else:
                    visibleCombo.setCurrentIndex(1)
                dlg.instanceLabelsLayout.addWidget(visibleCombo, row_index, 2)
                row_index += 1
        # now list instance attributes
        for counter, name in enumerate(item.symattrs.keys()):
            dlg.instanceAttributesLayout.addWidget(edf.boldLabel(name, dlg),
                                                   counter, 0)
            labelType = edf.longLineEdit()
            labelType.setReadOnly(True)
            labelNameEdit = edf.longLineEdit()
            labelNameEdit.setText(item.symattrs.get(name))
            labelNameEdit.setToolTip(f"{name} attribute (Read Only)")
            dlg.instanceAttributesLayout.addWidget(labelNameEdit, counter, 1)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            libraryName = dlg.libNameEdit.text().strip()
            cellName = dlg.cellNameEdit.text().strip()
            viewName = dlg.viewNameEdit.text().strip()
            instanceTuple = ddef.viewNameTuple(libraryName, cellName, viewName)
            location = QPoint(int(float(dlg.xLocationEdit.text().strip())),
                              int(float(dlg.yLocationEdit.text().strip())), )
            newInstance = self.instSymbol(instanceTuple, location)

            if newInstance:
                newInstance.instanceName = dlg.instNameEdit.text().strip()
                newInstance.angle = float(dlg.angleEdit.text().strip())
                newInstance.counter = item.counter

                tempDoc = QTextDocument()
                for i in range(dlg.instanceLabelsLayout.rowCount()):
                    # first create label name document with HTML annotations
                    label_item = dlg.instanceLabelsLayout.itemAtPosition(i, 0)
                    if label_item is None or label_item.widget() is None:
                        continue
                    tempDoc.setHtml(label_item.widget().text())
                    # now strip html annotations
                    tempLabelName = f"@{tempDoc.toPlainText().strip()}"
                    # check if label name is in label dictionary of item.
                    if newInstance.labels.get(tempLabelName):
                        # this is where the label value is set.
                        newInstance.labels[tempLabelName].labelValue = (
                            dlg.instanceLabelsLayout.itemAtPosition(i,
                                                                    1).widget().text())
                        visible = (dlg.instanceLabelsLayout.itemAtPosition(i,
                                                                           2).widget().currentText())
                        if visible == "True":
                            newInstance.labels[tempLabelName].labelVisible = True
                        else:
                            newInstance.labels[tempLabelName].labelVisible = False
                [labelItem.labelDefs() for labelItem in
                 newInstance.labels.values()]
                newInstance.setPos(self.snapToGrid(location - self.origin))
                newInstance.flipTuple = item.flipTuple
                self.undoStack.push(
                    us.addDeleteShapeUndo(self, newInstance, item))

    def setNetProperties(self, netItem: snet.schematicNet):
        dlg = pdlg.netProperties(self.editorWindow)
        dlg.netStartPointEditX.setText(
            str(round(netItem.mapToScene(netItem.draftLine.p1()).x())))
        dlg.netStartPointEditY.setText(
            str(round(netItem.mapToScene(netItem.draftLine.p1()).y())))
        dlg.netEndPointEditX.setText(
            str(round(netItem.mapToScene(netItem.draftLine.p2()).x())))
        dlg.netEndPointEditY.setText(
            str(round(netItem.mapToScene(netItem.draftLine.p2()).y())))
        dlg.netNameEdit.setText(netItem.name)
        dlg.widthButtonGroup.button(netItem.width).setChecked(True)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            netName = dlg.netNameEdit.text().strip()
            netStartX = float(dlg.netStartPointEditX.text())
            netStartY = float(dlg.netStartPointEditY.text())
            netStart = self.snapToGrid(QPoint(netStartX, netStartY))
            netEndX = float(dlg.netEndPointEditX.text())
            netEndY = float(dlg.netEndPointEditY.text())
            netEnd = self.snapToGrid(QPoint(netEndX, netEndY))
            newNet = snet.schematicNet(netStart, netEnd, netItem.mode)
            newNet.nameStrength = snet.netNameStrengthEnum.SET
            if self.isValidNetName(netName):
                newNet.name = netName
            else:
                self.logger.warning(f"{netName} is malformed, please correct.")
                newNet.name = netItem.name
            newNet.width = dlg.widthButtonGroup.checkedId()
            self.undoStack.push(us.addDeleteShapeUndo(self, newNet, netItem))

    def setTextProperties(self, item):
        dlg = pdlg.noteTextEdit(self.editorWindow)
        dlg.plainTextEdit.setText(item.textContent)
        dlg.familyCB.setCurrentText(item.fontFamily)
        dlg.fontStyleCB.setCurrentText(item.fontStyle)
        dlg.fontsizeCB.setCurrentText(item.textHeight)
        dlg.textAlignmCB.setCurrentText(item.textAlignment)
        dlg.textOrientCB.setCurrentText(item.textOrient)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            # item.prepareGeometryChange()
            start = item.mapToScene(item.start)
            newText = shp.text(start, dlg.plainTextEdit.toPlainText(),
                               dlg.familyCB.currentText(),
                               dlg.fontStyleCB.currentText(),
                               dlg.fontsizeCB.currentText(),
                               dlg.textAlignmCB.currentText(),
                               dlg.textOrientCB.currentText(), )
            self.rotateAnItem(start, newText, int(float(item.textOrient[1:])))
            self.undoStack.push(us.addDeleteShapeUndo(self, newText, item))
        return item

    def setSchematicPinProperties(self, item: shp.schematicPin):
        dlg = pdlg.schematicPinPropertiesDialog(self.editorWindow)
        dlg.pinName.setText(item.pinName)
        dlg.pinDir.setCurrentText(item.pinDir)
        dlg.pinType.setCurrentText(item.pinType)
        dlg.angleEdit.setText(str(item.angle))
        dlg.xlocationEdit.setText(str(item.mapToScene(item.start).x()))
        dlg.ylocationEdit.setText(str(item.mapToScene(item.start).y()))
        if dlg.exec() == QDialog.DialogCode.Accepted:
            pinName = dlg.pinName.text().strip()
            pinDir = dlg.pinDir.currentText()
            pinType = dlg.pinType.currentText()
            itemStartPos = QPoint(int(float(dlg.xlocationEdit.text().strip())),
                                  int(float(dlg.ylocationEdit.text().strip())), )
            start = self.snapToGrid(itemStartPos - self.origin)
            angle = float(dlg.angleEdit.text().strip())
            newPin = shp.schematicPin(QPoint(0, 0), pinName, pinDir, pinType)
            newPin.setPos(start)
            newPin.angle = angle
            self.undoStack.push(us.addDeleteShapeUndo(self, newPin, item))

    def hilightNets(self):
        """
        Show the connections the selected items.
        """
        try:
            self.highlightNets = bool(
                self.editorWindow.hilightNetAction.isChecked())
        except Exception as e:
            self.logger.error(e)

    def goDownHier(self):
        """
        Go down the hierarchy, opening the selected view.
        """
        try:
            selectedSymbol = [item for item in self.selectedItems() if
                              isinstance(item, shp.schematicSymbol)][0]

            if isinstance(selectedSymbol, shp.schematicSymbol):
                dlg = fd.goDownHierDialogue(self.editorWindow)
                libItem = libm.getLibItem(
                    self.editorWindow.libraryView.libraryModel,
                    selectedSymbol.libraryName, )
                cellItem = libm.getCellItem(libItem, selectedSymbol.cellName)
                viewNames = [cellItem.child(i).text() for i in
                             range(cellItem.rowCount()) if
                             "schematic" in cellItem.child(
                                 i).text() or "symbol" in cellItem.child(
                                 i).text() or "veriloga" in cellItem.child(
                                 i).text() or "spice" in cellItem.child(
                                 i).text() ]
                dlg.viewListCB.addItems(viewNames)
                if dlg.exec() == QDialog.DialogCode.Accepted:
                    selectedSymbol.setSelected(False)
                    self.saveSchematic(self.editorWindow.file)
                    viewItem = libm.getViewItem(
                        cellItem, dlg.viewListCB.currentText()
                    )
                    viewItemT = ddef.viewItemTuple(libItem, cellItem, viewItem)
                    openViewNameT = self.editorWindow.libraryView.openCellView(
                            viewItemT)

                    if viewItem.viewType == "schematic":
                        parentInstanceName = [labelItem.labelValue for labelItem in
                             selectedSymbol.labels.values() if
                             labelItem.labelType == "NLPLabel" and
                            labelItem.labelDefinition == "[@instName]"][0]
                        self.editorWindow.appMainW.openViews[
                            openViewNameT].centralW.scene.hierarchyTrail = (
                            f"{self.hierarchyTrail}{parentInstanceName}.")
                    if self.editorWindow.appMainW.openViews[openViewNameT]:
                        childWindow = self.editorWindow.appMainW.openViews[
                            openViewNameT]
                        childWindow.parentEditor = self.editorWindow
                        childWindow.parentObj = selectedSymbol
                        childWindowType = self.findEditorTypeString(childWindow)

                        if childWindowType == "symbolEditor":
                            childWindow.symbolToolbar.addAction(
                                childWindow.goUpAction)
                            if dlg.buttonId == 2:
                                childWindow.centralW.scene.readOnly = True
                        elif childWindowType == "schematicEditor":
                            childWindow.schematicToolbar.addAction(
                                childWindow.goUpAction)
                            if dlg.buttonId == 2:
                                childWindow.centralW.scene.readOnly = True


        except IndexError:
            pass

    def ignoreSymbol(self):
        if self.selectedItems() is not None:
            for item in self.selectedItems():
                if isinstance(item, shp.schematicSymbol):
                    item.netlistIgnore = not item.netlistIgnore
        else:
            self.logger.warning("No symbol selected")

    def renumberInstances(self):
        symbolList = [item for item in self.items() if
                      isinstance(item, shp.schematicSymbol)]

        for index, symbolInstance in enumerate(symbolList):
            symbolInstance.counter = index
            if symbolInstance.instanceName.startswith("I"):
                # Preserve array notation like I2<0:5>
                if "<" in symbolInstance.instanceName and ">" in symbolInstance.instanceName:
                    array_part = symbolInstance.instanceName[
                        symbolInstance.instanceName.find("<"):]
                    symbolInstance.instanceName = f"I{index}{array_part}"
                else:
                    symbolInstance.instanceName = f"I{index}"
                for label in symbolInstance.labels.values():
                    label.labelDefs()
        self.instanceCounter = index + 1
        self.saveSchematic(self.editorWindow.file)
        self.reloadScene()

    ### netlisting related methods
    def genSymbolPinNetMap(self, symbolItem: shp.schematicSymbol):
        # Process connected pins and collect unconnected ones in single pass
        unconnectedPins = []
        for pinName, pinItem in symbolItem.pins.items():
            netsConnectedToPin = [netItem for netItem in pinItem.collidingItems(
                Qt.IntersectsItemBoundingRect) if
                                  isinstance(netItem, snet.schematicNet)]

            if netsConnectedToPin:
                # all the nets connected to this pin should have the same name
                netName = netsConnectedToPin[0].name
                symbolItem.pinNetMap[pinName] = netName
                pinItem.connected = True
            else:
                pinItem.connected = False
                unconnectedPins.append(pinName)

        # Assign nets to unconnected pins
        for pin in unconnectedPins:
            symbolItem.pinNetMap[pin] = f"dnet{self.netCounter}"
            self.netCounter += 1
        
        # Handle pin ordering
        pinOrder = symbolItem.symattrs.get("pinOrder")
        if pinOrder:
            ordered_map = {}
            pinNames = (item.strip() for item in pinOrder.split(","))
            for pinName in pinNames:
                ordered_map[pinName] = symbolItem.pinNetMap[pinName]
            symbolItem.pinNetMap = ordered_map

    def _processNetGroup(self, namedNets: Set[snet.schematicNet],
                         remainingNets: Set[snet.schematicNet]):
        """Process a group of named nets and their connections."""
        while namedNets:
            net = namedNets.pop()
            connectedNets = self.findConnectedNets(net, remainingNets)
            for connectedNet in connectedNets:
                connectedNet.mergeNetName(net)
            remainingNets -= connectedNets

    def findConnectedNets(self, startNet: snet.schematicNet,
                          candidateNets: Set[snet.schematicNet]) -> Set[
        snet.schematicNet]:
        """Find all nets connected to startNet using DFS without mutating input."""
        connected = set()
        stack = [startNet]
        visited = {startNet}

        while stack:
            currentNet = stack.pop()
            for candidateNet in candidateNets:
                if candidateNet not in visited and self.checkNetConnect(
                        currentNet, candidateNet):
                    connected.add(candidateNet)
                    stack.append(candidateNet)
                    visited.add(candidateNet)

        return connected

    def nameSceneNets(self):
        """
        Name all nets in the scene.
        """
        netCounter = 0
        schematicSymbolSet = self.findSceneSymbolSet()
        sceneNetsSet: Set[snet.schematicNet] = self.findSceneNetsSet()

        # Clear existing names
        for netItem in sceneNetsSet:
            netItem.clearName()

        # Process in priority order
        result, globalNetsSet = self.findGlobalNets(schematicSymbolSet)
        if not result:
            self.logger.error("Net name conflict in global pins.")
            return
        sceneNetsSet -= globalNetsSet
        # now follow all the connected nets to globalNetsSet nets
        self._processNetGroup(globalNetsSet, sceneNetsSet)

        result, schemPinConNetsSet = self.findSchPinNets()
        if not result:
            self.logger.error("Net name conflict in schematic pins.")
            return
        sceneNetsSet -= schemPinConNetsSet
        self._processNetGroup(schemPinConNetsSet, sceneNetsSet)

        namedNetsSet = {net for net in sceneNetsSet if
                        net.nameStrength.value == 3}
        sceneNetsSet -= namedNetsSet
        self._processNetGroup(namedNetsSet, sceneNetsSet)

        # Auto-name remaining nets
        while sceneNetsSet:
            net = sceneNetsSet.pop()
            net.name = f"net{netCounter}"
            net.nameStrength = snet.netNameStrengthEnum.WEAK
            connectedNets = self.findConnectedNets(net, sceneNetsSet)
            for connectedNet in connectedNets:
                connectedNet.mergeNetName(net)
            sceneNetsSet -= connectedNets
            netCounter += 1

        self.netCounter = netCounter

    # Net finding methods
    def findGlobalNets(self, symbolSet: set[shp.schematicSymbol]) -> tuple[
        bool, set[snet.schematicNet]]:
        """
        This method finds all nets connected to global pins.
        """
        globalNetsSet = set()

        # Find all global pins in one pass
        globalPins = (pinItem for symbol in symbolSet for pinName, pinItem in
                      symbol.pins.items() if pinName.endswith("!"))

        for pinItem in globalPins:
            try:
                # Get nets connected to this global pin
                connectedNets = (netItem for netItem in
                                 pinItem.collidingItems(Qt.IntersectsItemShape) if
                                 isinstance(netItem, snet.schematicNet))

                for netItem in connectedNets:
                    if netItem.nameStrength.value == 3:  # Strong name
                        if netItem.name != pinItem.pinName:
                            netItem.nameConflict = True
                            self.logger.error(
                                f"Net name conflict at {pinItem.pinName} of "
                                f"{pinItem.parent.instanceName}.")
                            return False, set()
                        else:
                            globalNetsSet.add(netItem)
                    else:
                        globalNetsSet.add(netItem)
                        netItem.name = pinItem.pinName
                        netItem.nameStrength = snet.netNameStrengthEnum.SET

            except AttributeError as e:
                self.logger.error(f"Error processing global pin {pinItem}: {e}")
                return False, set()

        return True, globalNetsSet

    def findSchPins(self, net: snet.schematicNet) -> set[shp.schematicPin]:
        return {pinItem for pinItem in net.collidingItems(Qt.IntersectsItemShape)
                if isinstance(pinItem, shp.schematicPin)}

    def findSchPinNets(self) -> tuple[bool, set[snet.schematicNet]]:
        """Find nets connected to schematic pins."""
        connectedNetsSet = set()

        for sceneSchemPin in self.findSceneSchemPinsSet():
            try:
                # Use collision detection for better performance
                connectedNets = (netItem for netItem in
                                 sceneSchemPin.collidingItems(
                                     Qt.IntersectsItemShape) if
                                 isinstance(netItem, snet.schematicNet))

                # Parse pin name once
                pinBaseName, pinIndices = self.parseBusNotation(
                    sceneSchemPin.pinName)

                for netItem in connectedNets:
                    netBaseName, netIndices = self.parseBusNotation(netItem.name)
                    if netItem.nameStrength.value == 3:
                        if netBaseName != pinBaseName:
                            netItem.nameConflict = True
                            self.logger.error(
                                f"Net name conflict: net '{netItem.name}' vs pin '{sceneSchemPin.pinName}'")
                            return False, set()
                        else:  # names are the same, check if bus sizes are the same
                            if abs(pinIndices[0] - pinIndices[1]) != abs(
                                    netIndices[0] - netIndices[1]):
                                netItem.nameConflict = True
                                self.logger.error(
                                    f"Incompatible bus sizes: {sceneSchemPin.pinName} <-> {netItem.name}")
                                return False, set()
                            connectedNetsSet.add(netItem)
                    elif netItem.nameStrength.value != 3:
                        netItem.name = sceneSchemPin.pinName
                        netItem.nameStrength = snet.netNameStrengthEnum.INHERIT
                        connectedNetsSet.add(netItem)

            except (AttributeError, ValueError) as e:
                self.logger.error(
                    f"Error processing schematic pin {sceneSchemPin.pinName}: {e}")
                return False, set()
        return True, connectedNetsSet

    def _processNetPinConnection(self, net: snet.schematicNet, pinName: str) -> \
            set[snet.schematicNet]:
        """Common logic for processing net-pin connections."""
        if net.nameStrength.value == 3 and net.name != pinName:
            net.nameConflict = True
            self.logger.error(
                f"Net name conflict: net '{net.name}' vs pin '{pinName}'")
            return set()

        if net.nameStrength.value != 3:
            net.name = pinName
            net.nameStrength = snet.netNameStrengthEnum.INHERIT
        return {net}

    def checkErrors(self):
        """
        Check for errors in the schematic.
        """
        self.logger.info("Checking for errors...")
        symbolSet = self.findSceneSymbolSet()
        overlapExists, overlapRects = schk.checkSymbolOverlaps(symbolSet)
        if overlapExists:
            errorPen = QPen(Qt.red, 1, Qt.DashDotLine)
            for rect in overlapRects:
                self.overlapRectSet.add(self.addRect(rect, errorPen))

    def deleteErrors(self):
        self.logger.info("Deleting errors...")
        for rectItem in self.overlapRectSet:
            self.removeItem(rectItem)

    @staticmethod
    def checkNetConnect(netItem, otherNetItem):
        """
        Determine if a net is connected to another one. One net should end on the other net.
        """
        if netItem is otherNetItem:
            return False

        # Check if any endpoint pairs are within connection tolerance
        return any((netEnd - otherEnd).manhattanLength() <= 1 for netEnd in
                   netItem.sceneEndPoints for otherEnd in
                   otherNetItem.sceneEndPoints)

    @staticmethod
    def isValidNetName(text: str) -> bool:
        """
        Validates if the string is either:
        1. A string without any < or > characters
        2. A string ending with <int:int> where int is a positive integer

        Returns True if valid, False otherwise
        """
        # Pattern for string ending with <int:int>
        bus_pattern = QRegularExpression(r"^.*<(\d+):(\d+)>$")

        # Pattern to detect any partial bus notation
        partial_pattern = QRegularExpression(r"[<>:]")

        # If there are no <, >, or : characters, it's a valid simple string
        if not partial_pattern.match(text).hasMatch():
            return True

        # If it contains any of <, >, or :, check if it ends with complete pattern
        match = bus_pattern.match(text)
        if match.hasMatch() and text.count("<") == 1 and text.count(">") == 1:
            return True
        return False

    @staticmethod
    def createBusRanges(start: int, end: int):
        if start < end:
            resultRange = range(start, end + 1)
        else:
            resultRange = range(start, end - 1, -1)
        return resultRange

    @staticmethod
    def parseBusNotation(name: str) -> tuple[str, tuple[int, int]]:
        """
        Parse bus notation like 'name<0:5>' into base name and index range.
        Also handles single net notation like 'name<0>' or 'name<1>'.
    
        Args:
        name (str): The net name with optional bus notation.
    
        Returns:
        tuple[str, tuple[int, int]]: A tuple containing the base name and a tuple of start and end indices.
        """
        # Check if the name does not contain bus notation
        if '<' not in name or '>' not in name:
            return name, (0, 0)

        baseName = name.split('<')[0]  # Extract the base name before '<'
        indexRange = name.split('<')[1].split('>')[
            0]  # Extract the content inside '<>'

        # Check if it's a single index (e.g., 'name<0>')
        if ':' not in indexRange:
            singleIndex = int(indexRange)
            return baseName, (singleIndex, singleIndex)

        # Handle range notation (e.g., 'name<0:5>')
        start, end = map(int, indexRange.split(':'))
        return baseName, (start, end)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# properties dialogues for various editor functions

import pathlib

from PySide6.QtCore import (
    Qt,
)
from PySide6.QtGui import (
    QFontDatabase,
)
from PySide6.QtWidgets import (
    QWidget,
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QFormLayout,
    QDialogButtonBox,
    QButtonGroup,
    QTabWidget,
    QLineEdit,
    QLabel,
    QComboBox,
    QGroupBox,
    QRadioButton,
    QGridLayout,
    QTextEdit,
    QCheckBox,
    QTableWidget,
    QTableWidgetItem,
)

import revedaEditor.common.labels as lbl
import revedaEditor.common.net as net
import revedaEditor.common.shapes as shp
import revedaEditor.gui.editFunctions as edf


class rectPropertyDialog(QDialog):
    """
    Property dialog for symbol rectangles.
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setMinimumWidth(300)
        self.setWindowTitle("Rectangle Properties")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 20, 10, 20)
        self.rectWidthLine = edf.shortLineEdit()
        self.fLayout.addRow(QLabel("Width:"), self.rectWidthLine)
        self.rectHeightLine = edf.shortLineEdit()
        self.fLayout.addRow(QLabel("Height:"), self.rectHeightLine)
        self.rectLeftLine = edf.shortLineEdit()
        self.fLayout.addRow(QLabel("X Origin:"), self.rectLeftLine)
        self.rectTopLine = edf.shortLineEdit()
        self.fLayout.addRow(QLabel("Y Origin:"), self.rectTopLine)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class circlePropertyDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setMinimumWidth(300)
        self.setWindowTitle("Circle Properties")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 10, 10, 10)
        self.centerXEdit = edf.shortLineEdit()

        self.fLayout.addRow(QLabel("center x-coord:"), self.centerXEdit)
        self.centerYEdit = edf.shortLineEdit()

        self.fLayout.addRow(QLabel("center y-coord:"), self.centerYEdit)
        self.radiusEdit = edf.shortLineEdit()

        self.fLayout.addRow(QLabel("radius:"), self.radiusEdit)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class arcPropertyDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Arc Properties")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 10, 10, 10)
        # self.mainLayout.addWidget(self.arcTypeCombo)
        self.startXEdit = edf.shortLineEdit()
        self.fLayout.addRow(edf.boldLabel("X Origin:"), self.startXEdit)
        self.startYEdit = edf.shortLineEdit()
        self.fLayout.addRow(edf.boldLabel("Y Origin:"), self.startYEdit)
        self.widthEdit = edf.shortLineEdit()

        self.fLayout.addRow(edf.boldLabel("Width:"), self.widthEdit)
        self.heightEdit = edf.shortLineEdit()
        self.fLayout.addRow(edf.boldLabel("Height:"), self.heightEdit)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class linePropertyDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent

        self.setWindowTitle("Line Properties")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 10, 10, 10)
        self.startXLine = QLineEdit()

        self.fLayout.addRow(QLabel("Start (X):"), self.startXLine)
        self.startYLine = edf.shortLineEdit()
        self.fLayout.addRow(QLabel("Start (Y):"), self.startYLine)
        self.endXLine = edf.shortLineEdit()
        self.fLayout.addRow(QLabel("End (X):"), self.endXLine)
        self.endYLine = edf.shortLineEdit()
        self.fLayout.addRow(QLabel("End (Y):"), self.endYLine)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class pointsTableWidget(QTableWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setColumnCount(3)
        self.setHorizontalHeaderLabels(["Del.", 'X', "Y"])
        self.setColumnWidth(0, 8)
        self.setShowGrid(True)
        self.setGridStyle(Qt.SolidLine)


class symbolPolygonProperties(QDialog):
    def __init__(self, parent: QWidget, tupleList: list):
        super().__init__(parent)
        self.tupleList = tupleList
        self.setWindowTitle("Symbol Polygon Properties")
        self.setMinimumWidth(300)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout = QVBoxLayout()
        self.setLayout(mainLayout)
        self.tableWidget = pointsTableWidget(self)
        mainLayout.addWidget(self.tableWidget)
        mainLayout.addWidget(self.buttonBox)
        self.populateTable()

    def populateTable(self):
        self.tableWidget.setRowCount(len(self.tupleList) + 1)  # Add one extra row

        for row, item in enumerate(self.tupleList):
            self.addRow(row, item)

        # Add an empty row at the end
        self.addEmptyRow(len(self.tupleList))

        # Connect cellChanged signal to handle when the last row is edited
        self.tableWidget.cellChanged.connect(self.handleCellChange)

    def addRow(self, row, item):

        delete_checkbox = QCheckBox()
        self.tableWidget.setCellWidget(row, 0, delete_checkbox)

        self.tableWidget.setItem(row, 1, QTableWidgetItem(str(item[0])))
        self.tableWidget.setItem(row, 2, QTableWidgetItem(str(item[1])))
        delete_checkbox.stateChanged.connect(
            lambda state, r=row: self.deleteRow(r, state)
        )

    def addEmptyRow(self, row):

        # self.table_widget.insertRow(row)
        delete_checkbox = QCheckBox()
        self.tableWidget.setCellWidget(row, 0, delete_checkbox)
        delete_checkbox.stateChanged.connect(
            lambda state, r=row: self.deleteRow(r, state))

        self.tableWidget.setItem(row, 1, QTableWidgetItem(""))
        self.tableWidget.setItem(row, 2, QTableWidgetItem(""))

    def handleCellChange(self, row, column):
        if (
                row == self.tableWidget.rowCount() - 1
        ):  # Check if last row and tuple text column
            if self.tableWidget.item(row, 2) is not None:
                text1 = self.tableWidget.item(row, 1).text()
                text2 = self.tableWidget.item(row, 2).text()
                if text1 != "" and text2 != "":
                    self.tableWidget.insertRow(row + 1)
                    self.addEmptyRow(row + 1)

    def deleteRow(self, row, state):
        if state == 2:  # Checked state
            self.tableWidget.removeRow(row)


class createPinDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.setWindowTitle("Create Pin")
        self.setMinimumWidth(300)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.pinName = edf.shortLineEdit()
        self.pinName.setPlaceholderText("Pin Name")
        self.pinName.setToolTip("Enter pin name")
        self.fLayout.addRow(QLabel("Pin Name"), self.pinName)
        self.pinDir = QComboBox()
        self.pinDir.addItems(shp.symbolPin.pinDirs)
        self.pinDir.setToolTip("Select pin direction")
        self.fLayout.addRow(QLabel("Pin Direction"), self.pinDir)
        self.pinType = QComboBox()
        self.pinType.addItems(shp.symbolPin.pinTypes)
        self.pinType.setToolTip("Select pin type")
        self.fLayout.addRow(QLabel("Pin Type"), self.pinType)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class pinPropertyDialog(createPinDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Pin Properties")
        self.pinXLine = edf.shortLineEdit()
        self.pinXLine.setToolTip("X Coordinate")
        self.fLayout.addRow(QLabel("X:"), self.pinXLine)
        self.pinYLine = edf.shortLineEdit()
        self.pinYLine.setToolTip("Y Coordinate")
        self.fLayout.addRow(QLabel("Y:"), self.pinYLine)


class createSymbolLabelDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.setWindowTitle("Create Label")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.labelDefinition = QLineEdit()
        self.labelDefinition.setPlaceholderText("Label Definition")
        self.labelDefinition.setToolTip("Enter label Definition")
        self.fLayout.addRow(QLabel("Label Definition"), self.labelDefinition)
        self.labelHeightEdit = QLineEdit()
        self.labelHeightEdit.setPlaceholderText("Label Height")
        self.labelHeightEdit.setToolTip("Enter label Height")
        self.fLayout.addRow(QLabel("Label Height"), self.labelHeightEdit)
        self.labelAlignCombo = QComboBox()
        self.labelAlignCombo.addItems(lbl.symbolLabel.labelAlignments)
        self.fLayout.addRow(QLabel("Label Alignment"), self.labelAlignCombo)
        self.labelOrientCombo = QComboBox()
        self.labelOrientCombo.addItems(lbl.symbolLabel.labelOrients)
        self.fLayout.addRow(QLabel("Label Orientation"), self.labelOrientCombo)
        self.labelUseCombo = QComboBox()
        self.labelUseCombo.addItems(lbl.symbolLabel.labelUses)
        self.fLayout.addRow(QLabel("Label Use"), self.labelUseCombo)
        self.labelVisiCombo = QComboBox()
        self.labelVisiCombo.addItems(["Yes", "No"])
        self.fLayout.addRow(QLabel("Label Visible"), self.labelVisiCombo)
        self.mainLayout.addLayout(self.fLayout)
        self.labelTypeGroup = QGroupBox("Label Type")
        self.labelTypeLayout = QHBoxLayout()
        self.normalType = QRadioButton(lbl.symbolLabel.labelTypes[0])
        self.normalType.setChecked(True)
        self.NLPType = QRadioButton(lbl.symbolLabel.labelTypes[1])
        self.pyLType = QRadioButton(lbl.symbolLabel.labelTypes[2])
        self.labelTypeLayout.addWidget(self.normalType)
        self.labelTypeLayout.addWidget(self.NLPType)
        self.labelTypeLayout.addWidget(self.pyLType)
        self.labelTypeGroup.setLayout(self.labelTypeLayout)
        self.mainLayout.addWidget(self.labelTypeGroup)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class labelPropertyDialog(createSymbolLabelDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Label Properties")
        self.labelXLine = edf.shortLineEdit()
        self.labelXLine.setToolTip("X Coordinate")
        self.fLayout.addRow(QLabel("X:"), self.labelXLine)
        self.labelYLine = edf.shortLineEdit()
        self.labelYLine.setToolTip("Y Coordinate")
        self.fLayout.addRow(QLabel("Y:"), self.labelYLine)


class symbolLabelsDialogue(QDialog):
    """
    Dialog for changing symbol labels and attributes. Symbol properties... menu item.
    """

    def __init__(self, parent, items: list, attributes: list):
        super().__init__(parent)
        self.parent = parent
        self.items = items
        self.attributes = attributes
        self.setWindowTitle("Symbol Labels")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.mainLayout = QVBoxLayout()
        self.symbolPropsLayout = QGridLayout()
        self.symbolLabelsLayout = QGridLayout()
        self.symbolAttrsMethod()
        # Symbol Labels Layout
        self.symbolLabelsMethod()
        labelsGroup = QGroupBox("Symbol Labels")
        labelsGroup.setLayout(self.symbolLabelsLayout)
        self.mainLayout.addWidget(labelsGroup)
        self.mainLayout.addStretch(1)
        # self.mainLayout.addLayout(self.symbolLabelsLayout)
        propsGroup = QGroupBox("Symbol Attributes")
        propsGroup.setLayout(self.symbolPropsLayout)
        self.mainLayout.addWidget(propsGroup)
        self.mainLayout.addStretch(1)
        # self.mainLayout.addLayout(self.symbolPropsLayout)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def symbolAttrsMethod(self):
        self.attributeNameList = []
        # self.attributeTypeList = []
        self.attributeDefList = []
        # Symbol Properties
        self.symbolPropsLayout.addWidget(QLabel("Attribute Name"), 0, 0)
        # self.symbolPropsLayout.addWidget(QLabel("Type"), 0, 1)
        self.symbolPropsLayout.addWidget(QLabel("Definition"), 0, 1)
        i = 0
        for item in self.attributes:
            self.attributeNameList.append(edf.longLineEdit())
            self.attributeNameList[i].setText(item.name)
            self.symbolPropsLayout.addWidget(self.attributeNameList[i], i + 1, 0)
            self.attributeDefList.append(edf.longLineEdit())
            self.attributeDefList[i].setText(item.definition)
            self.symbolPropsLayout.addWidget(self.attributeDefList[i], i + 1, 1)
            i += 1

        self.attributeNameList.append(edf.longLineEdit())
        self.attributeDefList.append(edf.longLineEdit())
        self.symbolPropsLayout.addWidget(self.attributeNameList[-1], i + 2, 0)
        self.symbolPropsLayout.addWidget(self.attributeDefList[-1], i + 2, 1)
        self.attributeNameList[-1].setPlaceholderText("Enter Attribute Name")
        self.attributeDefList[-1].setToolTip("Enter Attribute Definition")
        self.attributeDefList[-1].editingFinished.connect(
            lambda: self.updateAttributeDef(i + 1)
        )

    def updateAttributeDef(self, i):
        i += 1
        self.attributeNameList.append(edf.longLineEdit())
        self.attributeDefList.append(edf.longLineEdit())
        self.symbolPropsLayout.addWidget(self.attributeNameList[-1], i + 1, 0)
        self.symbolPropsLayout.addWidget(self.attributeDefList[-1], i + 1, 1)
        self.attributeNameList[-1].setPlaceholderText("Enter Attribute Name")
        self.attributeDefList[-1].setToolTip("Enter Attribute Definition")
        self.attributeDefList[-1].editingFinished.connect(
            lambda: self.updateAttributeDef(i)
        )

    def symbolLabelsMethod(self):
        self.labelDefinitionList = []
        self.labelHeightList = []
        self.labelAlignmentList = []
        self.labelOrientationList = []
        self.labelUseList = []
        self.labelTypeList = []
        self.labelItemList = []
        i = 0
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Definition"), 0, 0)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Height"), 0, 1)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Alignment"), 0, 2)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Orientation"), 0, 3)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Use"), 0, 4)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Type"), 0, 5)
        for item in self.items:
            if type(item) == lbl.symbolLabel:
                i += 1
                self.labelItemList.append(item)
                self.labelDefinitionList.append(edf.longLineEdit())
                self.labelDefinitionList[-1].setText(item.labelDefinition)
                self.labelDefinitionList[-1].setReadOnly(True)
                self.symbolLabelsLayout.addWidget(self.labelDefinitionList[i - 1], i, 0)
                self.labelHeightList.append(edf.shortLineEdit())
                self.labelHeightList[-1].setText(str(item.labelHeight))
                self.symbolLabelsLayout.addWidget(self.labelHeightList[i - 1], i, 1)
                self.labelAlignmentList.append(QComboBox())
                self.labelAlignmentList[-1].addItems(lbl.symbolLabel.labelAlignments)
                self.labelAlignmentList[-1].setCurrentText(item.labelAlign)
                self.symbolLabelsLayout.addWidget(self.labelAlignmentList[-1], i, 2)
                self.labelOrientationList.append(QComboBox())
                self.labelOrientationList[-1].addItems(lbl.symbolLabel.labelOrients)
                self.labelOrientationList[-1].setCurrentText(item.labelOrient)
                self.symbolLabelsLayout.addWidget(self.labelOrientationList[-1], i, 3)
                self.labelUseList.append(QComboBox())
                self.labelUseList[-1].addItems(lbl.symbolLabel.labelUses)
                self.labelUseList[-1].setCurrentText(item.labelUse)
                self.symbolLabelsLayout.addWidget(self.labelUseList[-1], i, 4)
                self.labelTypeList.append(QComboBox())
                self.labelTypeList[-1].addItems(lbl.symbolLabel.labelTypes)
                self.labelTypeList[-1].setCurrentText(item.labelType)
                self.symbolLabelsLayout.addWidget(self.labelTypeList[-1], i, 5)
        if i == 0:  # no labels to edit
            self.symbolLabelsLayout.addWidget(QLabel("No symbol labels found."), 1, 0)


class instanceProperties(QDialog):
    def __init__(self, parent):
        # assert isinstance(instance, shp.symbolShape)
        super().__init__(parent)
        self.parent = parent
        self.initUI()

    def initUI(self):
        self.setWindowTitle("Instance Properties")
        tabWidget = QTabWidget()
        mainLayout = QVBoxLayout()
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        formLayout = QFormLayout()
        self.libNameEdit = edf.longLineEdit()
        self.libNameEdit.setReadOnly(True)
        self.libNameEdit.setToolTip("Library Name (Read Only)")
        formLayout.addRow(edf.boldLabel("Library Name", self), self.libNameEdit)
        self.cellNameEdit = edf.longLineEdit()
        self.cellNameEdit.setReadOnly(True)
        self.cellNameEdit.setToolTip("Cell Name (Read Only)")
        formLayout.addRow(edf.boldLabel("Cell Name", self), self.cellNameEdit)
        self.viewNameEdit = edf.longLineEdit()
        self.viewNameEdit.setReadOnly(True)
        self.viewNameEdit.setToolTip("View Name (Read Only)")
        formLayout.addRow(edf.boldLabel("View Name", self), self.viewNameEdit)
        self.instNameEdit = edf.longLineEdit()
        self.instNameEdit.setToolTip("Instance Name")
        formLayout.addRow(edf.boldLabel("Instance Name", self), self.instNameEdit)
        self.xLocationEdit = edf.shortLineEdit()
        formLayout.addRow(edf.boldLabel("x location", self), self.xLocationEdit)
        self.yLocationEdit = edf.shortLineEdit()
        formLayout.addRow(edf.boldLabel("y location", self), self.yLocationEdit)
        self.angleEdit = edf.longLineEdit()

        formLayout.addRow(edf.boldLabel("Angle", self), self.angleEdit)
        formLayout.setVerticalSpacing(10)
        mainLayout.addLayout(formLayout)
        propertiesGroup = QGroupBox()
        self.instanceLabelsLayout = QGridLayout()
        self.instanceLabelsLayout.setColumnMinimumWidth(0, 100)
        self.instanceLabelsLayout.setColumnMinimumWidth(1, 200)
        self.instanceLabelsLayout.setColumnMinimumWidth(2, 100)
        self.instanceLabelsLayout.setColumnStretch(0, 0)
        self.instanceLabelsLayout.setColumnStretch(1, 1)
        self.instanceLabelsLayout.setColumnStretch(2, 0)
        propertiesGroup.setLayout(self.instanceLabelsLayout)
        tabWidget.addTab(propertiesGroup, "Instance Properties")
        attributesGroup = QGroupBox()
        self.instanceAttributesLayout = QGridLayout()
        self.instanceAttributesLayout.setColumnMinimumWidth(0, 100)
        self.instanceAttributesLayout.setColumnMinimumWidth(1, 200)
        self.instanceAttributesLayout.setColumnStretch(0, 0)
        self.instanceAttributesLayout.setColumnStretch(1, 1)
        attributesGroup.setLayout(self.instanceAttributesLayout)
        tabWidget.addTab(attributesGroup, "Instance Attributes")
        mainLayout.addWidget(tabWidget)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)


class netNameDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Add Net Name")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        buttonBox = QDialogButtonBox(QBtn)
        mainLayout = QVBoxLayout()
        formLayout = QFormLayout()
        self.netNameEdit = edf.longLineEdit()
        self.netNameEdit.setToolTip("Enter net name")
        formLayout.addRow(edf.boldLabel("Net Name", self), self.netNameEdit)
        mainLayout.addLayout(formLayout)
        mainLayout.addSpacing(40)
        mainLayout.addWidget(buttonBox)
        self.setLayout(mainLayout)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)


class netProperties(QDialog):
    def __init__(self, parent):
        # assert isinstance(instance, shp.symbolShape)
        super().__init__(parent)
        self.parent = parent
        self.net = net
        self.initUI()

    def initUI(self):
        self.setWindowTitle("Net Properties")
        mainLayout = QVBoxLayout()
        okButton = QDialogButtonBox.Ok
        cancelButton = QDialogButtonBox.Cancel
        QBtn = okButton | cancelButton
        buttonBox = QDialogButtonBox(QBtn)
        buttonBox.button(QDialogButtonBox.Ok).setDefault(True)
        netPointsBox = QGroupBox("Net Points")
        netPointsLayout = QFormLayout()
        netPointsBox.setLayout(netPointsLayout)
        self.netStartPointEditX = edf.shortLineEdit()
        netPointsLayout.addRow(edf.boldLabel("Net Start X:"), self.netStartPointEditX)
        self.netStartPointEditY = edf.shortLineEdit()
        netPointsLayout.addRow(edf.boldLabel("Net Start Y:"), self.netStartPointEditY)
        self.netEndPointEditX = edf.shortLineEdit()
        netPointsLayout.addRow(edf.boldLabel("End Point X:"), self.netEndPointEditX)
        self.netEndPointEditY = edf.shortLineEdit()
        netPointsLayout.addRow(edf.boldLabel("End Point X:"), self.netEndPointEditY)

        netWidthGroupBox = QGroupBox('Net Width')
        netWidthLayout = QHBoxLayout()
        narrowNetButton = QRadioButton('Narrow')
        wideNetButton = QRadioButton('Wide')
        netWidthLayout.addWidget(narrowNetButton)
        netWidthLayout.addWidget(wideNetButton)
        self.widthButtonGroup = QButtonGroup(self)
        self.widthButtonGroup.addButton(narrowNetButton)
        self.widthButtonGroup.addButton(wideNetButton)
        self.widthButtonGroup.setId(narrowNetButton, 0)
        self.widthButtonGroup.setId(wideNetButton, 1)
        netWidthGroupBox.setLayout(netWidthLayout)
        mainLayout.addWidget(netWidthGroupBox)
        formBox = QGroupBox("Net Properties")
        formLayout = QFormLayout()
        self.netNameEdit = edf.longLineEdit()
        formLayout.addRow(edf.boldLabel("Net Name", self), self.netNameEdit)
        formBox.setLayout(formLayout)
        mainLayout.addWidget(formBox)
        mainLayout.addSpacing(20)
        mainLayout.addWidget(netPointsBox)
        mainLayout.addSpacing(40)
        mainLayout.addWidget(buttonBox)
        self.setLayout(mainLayout)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        self.show()


class createSchematicPinDialog(createPinDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Create Schematic Pin")


class schematicPinPropertiesDialog(createPinDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Pin Properties")
        self.xlocationEdit = edf.shortLineEdit()
        self.xlocationEdit.setToolTip("x location of pin")
        self.fLayout.addRow("x location:", self.xlocationEdit)
        self.ylocationEdit = edf.shortLineEdit()
        self.ylocationEdit.setToolTip("y location of pin")
        self.fLayout.addRow("y location:", self.ylocationEdit)
        self.angleEdit = edf.shortLineEdit()
        self.angleEdit.setToolTip("angle of pin")
        self.fLayout.addRow("angle:", self.angleEdit)


class symbolNameDialog(QDialog):
    def __init__(self, cellPath: pathlib.Path, cellName: str, parent):
        super().__init__(parent)
        self.cellPath = cellPath
        self.cellName = cellName
        self.symbolViewNames = [
            view.stem for view in cellPath.iterdir() if "symbol" in view.stem
        ]
        self.initUI()

    def initUI(self):
        self.setWindowTitle("Create a symbol?")
        self.mainLayout = QVBoxLayout()
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        formLayout = QFormLayout()
        formLayout.addRow(
            edf.boldLabel("Library Name"), QLabel(self.cellPath.parent.stem)
        )
        formLayout.addRow(edf.boldLabel("Cell Name"), QLabel(self.cellPath.stem))
        self.symbolViewsCB = QComboBox()
        self.symbolViewsCB.addItems(self.symbolViewNames)
        self.symbolViewsCB.setEditable(True)
        formLayout.addRow(edf.boldLabel("Symbol View Name:"), self.symbolViewsCB)
        self.mainLayout.addLayout(formLayout)
        self.mainLayout.addSpacing(40)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)


class symbolCreateDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Create Symbol")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.mainLayout = QVBoxLayout()

        self.fLayout = QFormLayout()
        self.topPinsEdit = edf.longLineEdit()
        self.topPinsEdit.setToolTip("Enter top pins")
        self.fLayout.addRow(edf.boldLabel("Top Pins:"), self.topPinsEdit)
        self.leftPinsEdit = edf.longLineEdit()
        self.leftPinsEdit.setToolTip("Enter left pins")
        self.fLayout.addRow(edf.boldLabel("Left Pins:"), self.leftPinsEdit)
        self.bottomPinsEdit = edf.longLineEdit()
        self.bottomPinsEdit.setToolTip("Enter bottom pins")
        self.fLayout.addRow(edf.boldLabel("Bottom Pins:"), self.bottomPinsEdit)
        self.rightPinsEdit = edf.longLineEdit()
        self.rightPinsEdit.setToolTip("Enter right pins")
        self.fLayout.addRow(edf.boldLabel("Right Pins:"), self.rightPinsEdit)
        self.mainLayout.addLayout(self.fLayout)
        self.mainLayout.addSpacing(20)
        self.geomLayout = QFormLayout()
        self.stubLengthEdit = QLineEdit()
        self.stubLengthEdit.setText("60")
        self.stubLengthEdit.setToolTip("Enter stub lengths")
        self.geomLayout.addRow(edf.boldLabel("Stub Length:"), self.stubLengthEdit)
        self.pinDistanceEdit = QLineEdit()
        self.pinDistanceEdit.setText("80")
        self.pinDistanceEdit.setToolTip("Enter pin spacing")
        self.geomLayout.addRow(edf.boldLabel("Pin spacing:"), self.pinDistanceEdit)
        self.mainLayout.addLayout(self.geomLayout)
        self.mainLayout.addSpacing(40)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class noteTextEdit(QDialog):
    """
    Set text properties.
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Edit Text")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        mainLayout = QVBoxLayout()
        self.plainTextEdit = QTextEdit()
        mainLayout.addWidget(self.plainTextEdit)
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [
            family for family in fontFamilies if QFontDatabase.isFixedPitch(family)
        ]
        formLayout = QFormLayout()
        self.familyCB = QComboBox()
        self.familyCB.addItems(fixedFamilies)
        self.familyCB.currentTextChanged.connect(self.familyFontStyles)
        formLayout.addRow(edf.boldLabel("Font Name"), self.familyCB)
        self.fontStyleCB = QComboBox()
        self.fontStyles = QFontDatabase.styles(fixedFamilies[0])
        self.fontStyleCB.addItems(self.fontStyles)
        self.fontStyleCB.currentTextChanged.connect(self.styleFontSizes)
        formLayout.addRow(edf.boldLabel("Font Style"), self.fontStyleCB)
        self.fontsizeCB = QComboBox()
        self.fontSizes = [
            str(size)
            for size in QFontDatabase.pointSizes(fixedFamilies[0], self.fontStyles[0])
        ]
        self.fontsizeCB.addItems(self.fontSizes)
        formLayout.addRow(edf.boldLabel("Font Size"), self.fontsizeCB)
        self.textAlignmCB = QComboBox()
        self.textAlignmCB.addItems(shp.text.textAlignments)
        formLayout.addRow(edf.boldLabel("Text Alignment"), self.textAlignmCB)
        self.textOrientCB = QComboBox()
        self.textOrientCB.addItems(shp.text.textOrients)
        formLayout.addRow(edf.boldLabel("Text Orientation"), self.textOrientCB)
        mainLayout.addLayout(formLayout)
        mainLayout.addSpacing(40)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)
        self.show()

    def familyFontStyles(self, s):
        self.fontStyleCB.clear()
        self.fontStyles = QFontDatabase.styles(self.familyCB.currentText())
        self.fontStyleCB.addItems(self.fontStyles)

    def styleFontSizes(self, s):
        self.fontsizeCB.clear()
        selectedFamily = self.familyCB.currentText()
        selectedStyle = self.fontStyleCB.currentText()
        self.fontSizes = [
            str(size)
            for size in QFontDatabase.pointSizes(selectedFamily, selectedStyle)
        ]
        self.fontsizeCB.addItems(self.fontSizes)


class displayConfigDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Display Options")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.vLayout = QVBoxLayout()
        gridValueGroup = QGroupBox("Grid Values")
        self.fLayout = QFormLayout()
        gridValueGroup.setLayout(self.fLayout)
        self.majorGridEntry = QLineEdit()
        self.majorGridEntry.setToolTip(
            "Enter Dot or Line Grid Spacing Value as a multiple of scene grid"
        )
        self.fLayout.addRow("GridSpacing:", self.majorGridEntry)
        self.snapGridEdit = QLineEdit()
        self.snapGridEdit.setToolTip(
            "Enter the Snap Grid Value as a multiple of scene grid"
        )
        self.fLayout.addRow("Snap Distance", self.snapGridEdit)

        gridTypeGroup = QGroupBox("Grid Type")
        gridTypeLayout = QHBoxLayout()
        self.dotType = QRadioButton("Dot Grid")
        self.dotType.setChecked(True)
        self.lineType = QRadioButton("Line Grid")
        self.noType = QRadioButton("No Grid")
        gridTypeLayout.addWidget(self.dotType)
        gridTypeLayout.addWidget(self.lineType)
        gridTypeLayout.addWidget(self.noType)
        gridTypeGroup.setLayout(gridTypeLayout)

        self.vLayout.addWidget(gridValueGroup)
        self.vLayout.addWidget(gridTypeGroup)
        self.vLayout.addStretch(1)

        self.vLayout.addWidget(self.buttonBox)
        self.setLayout(self.vLayout)
        self.show()


class layoutDisplayConfigDialog(displayConfigDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Layout Display Options")
        self.dbuEntry = edf.shortLineEdit()
        self.dbuEntry.setToolTip("Number of points per um")
        self.dbuEntry.setReadOnly(True)
        self.fLayout.insertRow(0, "Process Points per um:", self.dbuEntry)


class selectConfigDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Selection Options")
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        vLayout = QVBoxLayout()
        selectionTypeGroup = QGroupBox("Selection Type")
        selectionTypeLayout = QHBoxLayout()
        self.fullSelection = QRadioButton("Full")
        self.partialSelection = QRadioButton("Partial")
        selectionTypeLayout.addWidget(self.fullSelection)
        selectionTypeLayout.addWidget(self.partialSelection)
        selectionTypeGroup.setLayout(selectionTypeLayout)
        vLayout.addWidget(selectionTypeGroup)
        # snapDistanceGroup = QGroupBox("Snap Distance")
        # snapDistanceLayout = QFormLayout()
        # self.snapDistanceEntry = edf.shortLineEdit()
        # snapDistanceLayout.addRow("Snap Distance", self.snapDistanceEntry)
        # snapDistanceGroup.setLayout(snapDistanceLayout)
        # vLayout.addWidget(snapDistanceGroup)
        vLayout.addStretch(1)
        vLayout.addWidget(self.buttonBox)
        self.setLayout(vLayout)
        self.show()


class moveByDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.setWindowTitle("Move By...")
        self.setMinimumWidth(250)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout = QVBoxLayout()
        coordsGroup = QGroupBox("Move By")
        coordsLayout = QFormLayout()
        self.xEdit = edf.shortLineEdit()
        self.yEdit = edf.shortLineEdit()
        coordsLayout.addRow("Move By in X:", self.xEdit)
        coordsLayout.addRow("Move By in Y:", self.yEdit)
        coordsGroup.setLayout(coordsLayout)
        self.mainLayout.addWidget(coordsGroup)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)



#    "Commons Clause" License Condition v1.0
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

import json
import os
import re
import shutil
import tempfile
import urllib.request
import zipfile
from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QTableWidget,
    QTableWidgetItem,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QProgressBar,
    QTextEdit,
    QVBoxLayout,
    QWidget,
    QHeaderView,
    QApplication,
)


class LibraryRegistryWindow(QMainWindow):
    DEFAULT_REGISTRY = "https://raw.githubusercontent.com/eskiyerli/revolutionEDA_libraries/main/libraries.json"

    def __init__(
        self,
        parent=None,
        registry_url: str | None = None,
        libraries_dir: Path | None = None,
    ):
        super().__init__(parent)
        self.setWindowTitle("Revolution EDA Library Registry")
        self.resize(800, 450)

        self.registry_url = registry_url or self.DEFAULT_REGISTRY
        self.librariesDir = libraries_dir or Path.cwd()
        self.librariesDir = self.librariesDir.resolve()

        self.appMainW = QApplication.instance().appMainW

        self._initUI()
        self._registry = []
        self.fetch_registry()

    def _initUI(self):
        central = QWidget()
        self.setCentralWidget(central)
        main_l = QVBoxLayout(central)

        # Installation prefix section
        prefix_w = QWidget()
        prefix_l = QHBoxLayout(prefix_w)
        prefix_l.setContentsMargins(0, 0, 0, 0)
        prefix_l.addWidget(QLabel("Installation Prefix:"))
        self.prefix_edit = QLineEdit(str(self.librariesDir))
        prefix_l.addWidget(self.prefix_edit, 1)
        self.browse_btn = QPushButton("Browse...")
        prefix_l.addWidget(self.browse_btn)
        main_l.addWidget(prefix_w)

        # Main content area
        content_w = QWidget()
        content_l = QHBoxLayout(content_w)
        content_l.setContentsMargins(0, 0, 0, 0)

        # Left panel - table
        left_w = QWidget()
        left_l = QVBoxLayout(left_w)
        left_l.setContentsMargins(0, 0, 0, 0)
        self.tableWidget = QTableWidget()
        self.tableWidget.setColumnCount(5)
        self.tableWidget.setHorizontalHeaderLabels(
            ["Installed", "Library", "Type", "Version", "License"]
        )
        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        self.tableWidget.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        self.tableWidget.setSelectionMode(QTableWidget.SelectionMode.SingleSelection)
        left_l.addWidget(QLabel("Available libraries:"))
        left_l.addWidget(self.tableWidget)
        self.refresh_btn = QPushButton("Refresh")
        left_l.addWidget(self.refresh_btn)
        content_l.addWidget(left_w, 2)

        # Right panel - details
        right_w = QWidget()
        right_l = QVBoxLayout(right_w)
        right_l.setContentsMargins(0, 0, 0, 0)
        right_l.addWidget(QLabel("Description:"))
        self.desc = QTextEdit()
        self.desc.setReadOnly(True)
        right_l.addWidget(self.desc, 1)
        self.progress = QProgressBar()
        self.progress.setRange(0, 100)
        right_l.addWidget(self.progress)
        self.download_btn = QPushButton("Download / Install")
        self.uninstall_btn = QPushButton("Uninstall")
        right_l.addWidget(self.download_btn)
        right_l.addWidget(self.uninstall_btn)
        content_l.addWidget(right_w, 2)

        main_l.addWidget(content_w)

        # Connect signals
        self.tableWidget.itemSelectionChanged.connect(self._onSelect)
        self.tableWidget.itemDoubleClicked.connect(self._on_item_activated)
        self.download_btn.clicked.connect(self._on_download)
        self.uninstall_btn.clicked.connect(self._on_uninstall)
        self.refresh_btn.clicked.connect(self.fetch_registry)
        self.browse_btn.clicked.connect(self._on_browse)
        self.prefix_edit.textChanged.connect(self._on_prefix_changed)

    def _on_browse(self):
        directory = QFileDialog.getExistingDirectory(
            self, "Select Installation Directory", str(self.librariesDir)
        )
        if directory:
            self.prefix_edit.setText(directory)

    def _on_prefix_changed(self, text: str):
        self.librariesDir = Path(text).resolve()
        self.fetch_registry()

    def fetch_registry(self):
        self.tableWidget.setRowCount(0)
        self.desc.clear()
        self.progress.setValue(0)
        try:
            with urllib.request.urlopen(self.registry_url) as resp:
                raw = resp.read()
                index = json.loads(raw.decode("utf-8"))
                if isinstance(index, dict) and "libraries" in index:
                    index = index["libraries"]
                self._registry = index if isinstance(index, list) else []
        except Exception as exc:
            QMessageBox.warning(
                self, "Registry Error", f"Failed to fetch registry:\n{exc}"
            )
            self._registry = []

        self.tableWidget.setRowCount(len(self._registry))
        for row, entry in enumerate(self._registry):
            name = entry.get("name", "unknown")
            library_dir = self.librariesDir / re.sub(r"[^A-Za-z0-9_.-]", "_", name)

            checkbox_item = QTableWidgetItem()
            checkbox_item.setFlags(Qt.ItemFlag.ItemIsEnabled)
            checkbox_item.setCheckState(
                Qt.CheckState.Checked
                if library_dir.exists()
                else Qt.CheckState.Unchecked
            )
            checkbox_item.setData(Qt.ItemDataRole.UserRole, entry)
            self.tableWidget.setItem(row, 0, checkbox_item)

            self.tableWidget.setItem(row, 1, QTableWidgetItem(name))
            self.tableWidget.setItem(
                row, 2, QTableWidgetItem(entry.get("type", "design").title())
            )
            self.tableWidget.setItem(
                row, 3, QTableWidgetItem(entry.get("version", "0.0.0"))
            )
            self.tableWidget.setItem(
                row, 4, QTableWidgetItem(entry.get("license", "Unknown"))
            )

    def _onSelect(self):
        self.desc.clear()
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            return
        item = self.tableWidget.item(current_row, 0)
        if not item:
            return
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        text = f"{entry.get('description', '')}\n\nType: {entry.get('type', 'design').title()}\nVersion: {entry.get('version', 'N/A')}\nLicense: {entry.get('license', 'Unknown')}\nURL: {entry.get('url', '')}"
        self.desc.setPlainText(text)

    def _on_item_activated(self, item: QTableWidgetItem):
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        name = entry.get("name", "library")
        ret = QMessageBox.question(
            self,
            "Install Library",
            f"Install library '{name}'?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if ret == QMessageBox.StandardButton.Yes:
            self._install_entry(entry)

    def _on_download(self):
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            QMessageBox.information(
                self, "No selection", "Please select a library first."
            )
            return
        item = self.tableWidget.item(current_row, 0)
        if item:
            entry = item.data(Qt.ItemDataRole.UserRole) or {}
            name = entry.get("name", "library")
            ret = QMessageBox.question(
                self,
                "Install Library",
                f"Install '{name}'?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if ret == QMessageBox.StandardButton.Yes:
                self._install_entry(entry)

    def _on_uninstall(self):
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            QMessageBox.information(
                self, "No selection", "Please select a library first."
            )
            return
        item = self.tableWidget.item(current_row, 0)
        if not item:
            return
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        name = re.sub(r"[^A-Za-z0-9_.-]", "_", entry.get("name", "library"))
        library_dir = self.librariesDir / name

        if not library_dir.exists():
            QMessageBox.information(self, "Not Installed", f"'{name}' is not installed.")
            return

        ret = QMessageBox.question(
            self,
            "Uninstall Library",
            f"Uninstall '{name}'?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if ret == QMessageBox.StandardButton.Yes:
            try:
                shutil.rmtree(library_dir, ignore_errors=True)
                QMessageBox.information(self, "Success", f"'{name}' uninstalled.")
                self.fetch_registry()
            except Exception as e:
                QMessageBox.critical(self, "Error", f"Failed to uninstall:\n{e}")

    def _install_entry(self, entry: dict):
        name = re.sub(r"[^A-Za-z0-9_.-]", "_", entry.get("name", "library"))
        url = entry.get("url")

        if not url:
            QMessageBox.warning(self, "Error", "No URL available.")
            return

        target_subdir = self.librariesDir / name
        if target_subdir.exists():
            ok = QMessageBox.question(
                self,
                "Overwrite",
                f"{name} exists. Overwrite?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if ok != QMessageBox.StandardButton.Yes:
                return
            shutil.rmtree(target_subdir, ignore_errors=True)

        try:
            self.progress.setValue(10)
            with urllib.request.urlopen(url) as resp:
                tmp_fd, tmp_path = tempfile.mkstemp()
                os.close(tmp_fd)
                with open(tmp_path, "wb") as out:
                    out.write(resp.read())

            self.progress.setValue(50)
            self.librariesDir.mkdir(parents=True, exist_ok=True)
            if url.lower().endswith(".zip"):
                with zipfile.ZipFile(tmp_path, "r") as zf:
                    zf.extractall(path=self.librariesDir)
            else:
                target_subdir.mkdir(parents=True, exist_ok=True)
                (target_subdir / Path(url).name).write_bytes(
                    Path(tmp_path).read_bytes()
                )

            os.remove(tmp_path)
            self.progress.setValue(100)
            QMessageBox.information(self, "Success", f"'{name}' installed successfully.")
            self.fetch_registry()
            self.appMainW.libraryDict[name]=target_subdir
            self.appMainW.libraryBrowser.writeLibDefFile(self.appMainW.libraryDict,
                                                         self.appMainW.libraryBrowser.libFilePath)
            self.appMainW.libraryBrowser.designView.reworkDesignLibrariesView(
                self.appMainW.libraryDict)
        except Exception as e:
            self.progress.setValue(0)
            QMessageBox.critical(self, "Error", f"Installation failed:\n{e}")



#    "Commons Clause" License Condition v1.0
#
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, "Sell" means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)


from collections import deque
from dataclasses import dataclass
from typing import List

from PySide6.QtCore import QObject, QProcess, Slot
from PySide6.QtGui import QWindow


@dataclass
class Process:
    program: str
    arguments: List
    outputFile: str = ""
    process: QProcess = None
    id: int = 0


class ProcessManager(QObject):
    def __init__(self, max_processes=3, parent=None):
        super().__init__()
        self._maxProcesses = max_processes
        self.parent = parent
        self.process_queue = deque()
        self.running_processes = []
        self.processCounter = 0

    def add_process(self, program, arguments, outputFile=None):
        # print(f'program: {program}, arguments: {arguments}, outputFile: {outputFile}')
        self.processCounter += 1
        if outputFile:
            processItem = Process(program=program, arguments=arguments,
                                  outputFile=outputFile)
        else:
            processItem = Process(program=program, arguments=arguments)
        self.process_queue.append(processItem)
        self.start_next_process()
        return processItem

    def start_next_process(self):
        if len(self.running_processes) < self._maxProcesses and self.process_queue:
            processItem = self.process_queue.popleft()
            processItem.process = QProcess(self)
            processItem.process.setProgram(processItem.program)
            processItem.process.setArguments(processItem.arguments)
            if processItem.outputFile:
                processItem.process.setStandardOutputFile(processItem.outputFile)
            processItem.id = f'{processItem.arguments[-1]}_{self.processCounter}'
            processItem.process.readyReadStandardOutput.connect(lambda: self.handle_output(
                processItem.process))
            processItem.process.finished.connect(self.process_finished)
            processItem.process.start()
            self.running_processes.append(processItem)

    @Slot(int, QProcess.ExitStatus)
    def process_finished(self, exit_code, exit_status):
        finishedProcess = self.sender()
        processItem = next(
            (item for item in self.running_processes if item.process == finishedProcess),
            None)
        self.running_processes.remove(processItem)
        processItem.process.deleteLater()
        self.start_next_process()
        self.parent.logger.info(f"Process {processItem.id} finished.")

    @Slot(QProcess)
    def handle_output(self, process):
        output = process.readAllStandardOutput().data().decode("utf-8").strip()
        for line in output.split("\n"):
            self.parent.logger.info(line)

    def stop_all(self):
        for processItem in self.running_processes:
            processItem.process.terminate()
        self.process_queue.clear()

    @property
    def maxProcesses(self):
        return self._maxProcesses

    @maxProcesses.setter
    def maxProcesses(self, value: int):
        if isinstance(value, int):
            self._maxProcesses = value


class ProcessRunner():
    def __init__(self, parent: QWindow, program: str, arguments: List[str]):
        self.process = QProcess()
        self.process.readyReadStandardOutput.connect(lambda: self.handle_output)
        self.process.finished.connect(self.process_finished)
        self.process.readyReadStandardError.connect(lambda: self.handle_error)
        self.process.setProgram(program)
        self.process.setArguments(arguments)
        self.parent = parent
        self.process.start()
        self.process.waitForFinished()

    def handle_output(self, process):
        output = process.readAllStandardOutput().data().decode("utf-8").strip()
        for line in output.split("\n"):
            self.parent.logger.info(line)

    def handle_error(self, process):
        output = process.readAllStandardError().data().decode("utf-8").strip()
        for line in output.split("\n"):
            self.parent.logger.info(line)

    def process_finished(self, exit_code, exit_status):
        self.parent.logger.info(
            f"Process finished with exit code {exit_code} and status {exit_status}")



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
from collections import Counter

from PySide6.QtCore import (
    QLine,
    QPoint,
    Qt,
    Signal,
)
from PySide6.QtGui import (
    QColor,
    QKeyEvent,
    QPainter,
    QPolygon,
    QWheelEvent,
)
from PySide6.QtOpenGLWidgets import QOpenGLWidget
from PySide6.QtWidgets import (QGraphicsView,)

import revedaEditor.common.net as net
from revedaEditor.backend.pdkLoader import importPDKModule
from revedaEditor.scenes.editorScene import editorScene
from revedaEditor.scenes.layoutScene import layoutScene
from revedaEditor.scenes.schematicScene import schematicScene
from revedaEditor.scenes.symbolScene import symbolScene

schlyr = importPDKModule("schLayers")
fabproc = importPDKModule("process")


class editorView(QGraphicsView):
    """
    The qgraphicsview for qgraphicsscene. It is used for both schematic and layout editors.
    """

    keyPressedSignal = Signal(int)

    # zoomFactorChanged = Signal(float)
    def __init__(self, scene: editorScene, parent):
        super().__init__(scene, parent)

        # Cache parentW references
        self.centralW = parent
        self.editorWindow = self.centralW.editorWindow
        self.viewScene = scene
        self.logger = scene.logger

        # Cache editor properties
        editor = self.editorWindow
        self.majorGrid = editor.majorGrid
        self.snapGrid = editor.snapGrid
        self.snapTuple = editor.snapTuple

        # Direct attribute initialization
        self.gridbackg = True
        self.linebackg = self._transparent = False
        self.zoomFactor = 1.0

        # Initialize coordinate cache as integers (faster than QPoint)
        self._left = self._right = self._top = self._bottom = 0

        # Defer expensive operations
        self.viewRect = None
        self.init_UI()
        glWidget = QOpenGLWidget()
        glWidget.setUpdateBehavior(QOpenGLWidget.UpdateBehavior.PartialUpdate)
        self.setViewport(glWidget)


    def init_UI(self):
        """
        Initializes the user interface.
        """
        # Batch all settings to minimize Qt overhead
        self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.SmartViewportUpdate)
        self.setCacheMode(QGraphicsView.CacheModeFlag.CacheBackground)
        self.setMouseTracking(True)
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.setResizeAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.setInteractive(True)
        self.setCursor(Qt.CursorShape.CrossCursor)
        self.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.TextAntialiasing)


    def wheelEvent(self, event: QWheelEvent) -> None:
        """
        Handle the wheel event for zooming in and out of the view.

        Args:
            event (QWheelEvent): The wheel event to handle.
        """
        # Get the current center point of the view
        oldPos = self.mapToScene(self.viewport().rect().center())

        # Perform the zoom
        self.zoomFactor = 1.2 if event.angleDelta().y() > 0 else 1 / 1.2
        self.scale(self.zoomFactor, self.zoomFactor)

        # Get the new center point of the view
        newPos = self.mapToScene(self.viewport().rect().center())

        # Calculate the delta and adjust the scene position
        delta = newPos - oldPos
        self.translate(delta.x(), delta.y())
        self.determineViewRect()
    
    def determineViewRect(self):
        self.viewRect = self.mapToScene(self.rect()).boundingRect().toRect()

    def drawBackground(self, painter, rect):
        """
        Draws the background of the painter within the given rectangle.

        Args:
            painter (QPainter): The painter object to draw on.
            rect (QRect): The rectangle to draw the background within.
        """
        # Cache rect values to avoid multiple calls
        left = int(rect.left())
        top = int(rect.top())

        # Calculate coordinates once
        self._left = left - (left % self.majorGrid)
        self._top = top - (top % self.majorGrid)
        self._bottom = int(rect.bottom())
        self._right = int(rect.right())

        if self.gridbackg or self.linebackg:
            # Fill rectangle with black color
            painter.fillRect(rect, QColor("black"))
            x_coords, y_coords = self.findCoords()

            if self.gridbackg:
                painter.setPen(QColor("gray"))

                # Pre-allocate the polygon for better performance
                points = QPolygon()
                num_points = len(x_coords) * len(y_coords)
                points.reserve(num_points)

                # Fill the polygon with points
                for x in x_coords:
                    for y in y_coords:
                        points.append(QPoint(int(x), int(y)))

                # Draw all points in a single call
                painter.drawPoints(points)

            else:  # self.linebackg
                painter.setPen(QColor("gray"))

                # Create vertical and horizontal lines
                vertical_lines = [
                    QLine(int(x), self._top, int(x), self._bottom) for x in
                    x_coords
                ]

                horizontal_lines = [
                    QLine(self._left, int(y), self._right, int(y)) for y in
                    y_coords
                ]

                # Draw all lines with minimal calls
                painter.drawLines(vertical_lines)
                painter.drawLines(horizontal_lines)
        elif self._transparent:
            self.viewport().setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        else:
            painter.fillRect(rect, QColor("black"))
            super().drawBackground(painter, rect)

    def findCoords(self):
        """
        Calculate the coordinates for drawing lines or points on a grid.

        Returns:
            tuple: A tuple containing the x and y coordinates for drawing the lines or points.
        """
        x_coords = range(self._left, self._right, self.majorGrid)
        y_coords = range(self._top, self._bottom, self.majorGrid)

        num_lines = len(x_coords)
        if 120 <= num_lines < 240:
            spacing = self.majorGrid * 2
        elif 240 <= num_lines < 480:
            spacing = self.majorGrid * 4
        elif 480 <= num_lines < 960:
            spacing = self.majorGrid * 8
        elif 960 <= num_lines < 1920:
            spacing = self.majorGrid * 16
        elif num_lines >= 1920:
            return range(0, 0), range(0, 0)  # No grid when too dense
        else:
            spacing = self.majorGrid

        if spacing != self.majorGrid:
            x_coords = range(self._left, self._right, spacing)
            y_coords = range(self._top, self._bottom, spacing)

        return x_coords, y_coords

    def keyPressEvent(self, event: QKeyEvent):
        self.keyPressedSignal.emit(event.key())

        match event.key():
            case Qt.Key.Key_M:
                self.viewScene.editModes.setMode("moveItem")
                self.editorWindow.messageLine.setText("Move Item")
            case Qt.Key.Key_F:
                self.viewScene.fitItemsInView()
                self.editorWindow.messageLine.setText("Fit Items In View")
            case Qt.Key.Key_Left:
                self.viewScene.moveSceneLeft()
                self.editorWindow.messageLine.setText("Move View Left")
            case Qt.Key.Key_Right:
                self.viewScene.moveSceneRight()
                self.editorWindow.messageLine.setText("Move View Right")
            case Qt.Key.Key_Up:
                self.viewScene.moveSceneUp()
                self.editorWindow.messageLine.setText("Move View Up")
            case Qt.Key.Key_Down:
                self.viewScene.moveSceneDown()
                self.editorWindow.messageLine.setText("Move View Down")
            case Qt.Key.Key_PageUp:
                self.cycleSelection(event)
            case Qt.Key.Key_Z:
                if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                    self.viewScene.zoomInBy2()
                if event.modifiers() & Qt.KeyboardModifier.ShiftModifier:
                    self.viewScene.zoomOutBy2()
                else:
                    self.viewScene.editModes.setMode("zoomView")

            case Qt.Key.Key_Escape:
                self.clearStretchItems()
                self.viewScene.deselectAll()
                self.viewScene.selectedItemsSet = set()
                if self.viewScene.selectionRectItem:
                    self.viewScene.removeItem(self.viewScene.selectionRectItem)
                    self.viewScene.selectionRectItem = None
                if self.viewScene.selectedItemGroup:
                    self.viewScene.destroyItemGroup(
                        self.viewScene.selectedItemGroup)
                    self.viewScene.selectedItemGroup = None
                self.viewScene.editModes.setMode('selectItem')
                self.viewScene.messageLine.setText("Select Item")
            case _:
                super().keyPressEvent(event)

    def printView(self, printer):
        """
        Print view using selected Printer.

        Args:
            printer (QPrinter): The printer object to use for printing.

        This method prints the current view using the provided printer. It first creates a QPainter object
        using the printer. Then, it stores the original states of gridbackg and linebackg attributes.
        After that, it calls the revedaPrint method to render the view onto the painter. Finally, it
        restores the gridbackg and linebackg attributes to their original state.
        """
        # Store original states
        originalGridbackg = self.gridbackg
        originalLinebackg = self.linebackg

        # Set both to False for printing
        self.gridbackg = False
        self.linebackg = False
        self._transparent = True
        painter = QPainter(printer)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setRenderHint(QPainter.RenderHint.TextAntialiasing, True)
        painter.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform, True)
        targetRect = painter.viewport()
        sourceRect = self.sceneRect()
        self.scene().render(painter, targetRect, sourceRect)
        # Restore original states
        self.gridbackg = originalGridbackg
        self.linebackg = originalLinebackg
        self._transparent = False
        painter.end()

    def cycleSelection(self, event):
        if self.viewScene.itemCycler:
            item = next(self.viewScene.itemCycler)
            # Deselect without calling deselectAll(), which would reset itemCycler.
            for i in self.viewScene.selectedItems():
                i.setSelected(False)
            self.viewScene.clearSelection()
            self.viewScene.selectedItemsSet = {item}
            item.setSelected(True)

    def clearStretchItems(self):
        self.determineViewRect()
        for item in self.viewScene.items(self.viewRect):
            if hasattr(item, "stretch"):
                setattr(item, "stretch", False)

class symbolView(editorView):
    def __init__(self, scene: symbolScene, parent):
        super().__init__(scene, parent)
        self.viewScene: symbolScene = scene
        self.parent = parent

    def keyPressEvent(self, event: QKeyEvent):
        super().keyPressEvent(event)
        match event.key():
            case Qt.Key.Key_Escape:
                if self.viewScene.polygonGuideLine:
                    self.viewScene.finishPolygon(event)
                self.viewScene.newLine = None
                self.viewScene.newCircle = None
                self.viewScene.newPin = None
                self.viewScene.newRect = None
                self.viewScene.newArc = None
                self.viewScene.newLabel = None
                if self.viewScene.polygonGuideLine:
                    self.viewScene.removeItem(self.viewScene.polygonGuideLine)
                self.viewScene.editModes.setMode("selectItem")


class schematicView(editorView):
    _dotRadius = 10

    def __init__(self, scene: schematicScene, parent):
        super().__init__(scene, parent)
        self.viewScene: schematicScene = scene
        self.parent = parent

    def mousePressEvent(self, event):
        self.viewRect = self.mapToScene(self.rect()).boundingRect().toRect()

        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self.viewRect = self.mapToScene(self.rect()).boundingRect().toRect()
        super().mouseReleaseEvent(event)
        if not (self.viewScene.editModes.drawWire or self.viewScene.editModes.drawBus):
            self.pruneShortNets()
            self.mergeSplitViewNets()

    def pruneShortNets(self):
        """Remove nets shorter than snap spacing."""
        snapSpacing = self.viewScene.snapTuple[0]
        netsInView = [
            netItem
            for netItem in self.viewScene.items(self.viewRect)
            if isinstance(netItem, net.schematicNet)
        ]
        for netItem in netsInView:
            if netItem.scene() and netItem.draftLine.length() < snapSpacing:
                self.viewScene.removeItem(netItem)

    def mergeSplitViewNets(self):
        netsInView = (
            netItem
            for netItem in self.viewScene.items(self.viewRect)
            if isinstance(netItem, net.schematicNet)
        )
        for netItem in netsInView:
            if netItem.scene():
                self.viewScene.mergeSplitNets(netItem)

    def drawBackground(self, painter, rect):
        super().drawBackground(painter, rect)

        # Early exit for large views to avoid performance issues
        if (self._right - self._left) > 2000 and (
                self._bottom - self._top) > 2000:
            return

        # Get nets in view with type filtering
        netsInView = [
            item
            for item in self.viewScene.items(rect)
            if isinstance(item, net.schematicNet)
        ]

        if not netsInView:
            return

        # Collect and count endpoints in one pass
        pointCounts = Counter()
        for netItem in netsInView:
            pointCounts.update(netItem.sceneEndPoints)

        # Filter junction points (count >= 3) and draw
        junctionPoints = [point for point, count in pointCounts.items() if
                          count >= 3]

        if junctionPoints:
            painter.setPen(schlyr.wirePen)
            painter.setBrush(schlyr.wireBrush)
            for point in junctionPoints:
                painter.drawEllipse(point, self._dotRadius, self._dotRadius)

    def keyPressEvent(self, event: QKeyEvent):
        """
        Handles the key press event for the editor view.

        Args:
            event (QKeyEvent): The key press event to handle.

        """
        if event.key() == Qt.Key.Key_Escape:
            # Esc key pressed, remove snap rect and reset states
            if self.viewScene._snapPointRect is not None:
                self.viewScene._snapPointRect.setVisible(False)
            if self.viewScene._newNet is not None:
                self.viewScene.wireEditFinished.emit(self.viewScene._newNet)
                self.viewScene._newNet = None
            elif self.viewScene._stretchNet is not None:
                # Stretch net mode, cancel stretch
                self.viewScene._stretchNet.setSelected(False)
                self.viewScene._stretchNet.stretch = False
                self.viewScene.mergeSplitNets(self.viewScene._stretchNet)
                self.viewScene._stretchNet = None
            self.viewScene._newInstance = None
            self.viewScene._newPin = None
            self.viewScene._newText = None
            # Set the edit mode to select item
            self.viewScene.editModes.setMode("selectItem")

        super().keyPressEvent(event)


class layoutView(editorView):
    def __init__(self, scene: layoutScene, parent):
        super().__init__(scene, parent)
        self.viewScene: layoutScene = scene
        self.parent = parent
        # OpenGL viewport is already configured by editorView.__init__

    def keyPressEvent(self, event: QKeyEvent):
        modifiers = event.modifiers()
        if event.key() == Qt.Key.Key_Escape:
            if self.viewScene.editModes.drawPath and self.viewScene.newPath is not None:
                if self.viewScene.newPath.draftLine.isNull():
                    self.viewScene.undoStack.removeLastCommand()
                self.viewScene.newPath = None
                self.viewScene.editModes.setMode("selectItem")
            elif self.viewScene.editModes.drawRect and self.viewScene.newRect is not None:
                if self.viewScene.newRect.rect.isNull():
                    self.viewScene.removeItem(self.viewScene.newRect)
                    self.viewScene.undoStack.removeLastCommand()
                self.viewScene.newRect = None
                self.viewScene.editModes.setMode("selectItem")
            elif self.viewScene.editModes.stretchItem and self.viewScene.stretchPathItem is not None:
                self.viewScene.stretchPathItem.setSelected(False)
                self.viewScene.stretchPathItem.stretch = False
                self.viewScene.stretchPathItem = None
            elif self.viewScene.editModes.drawPolygon:
                if self.viewScene.polygonGuideLine:
                    self.viewScene.removeItem(self.viewScene.polygonGuideLine)
                if self.viewScene.newPolygon and self.viewScene.newPolygon.points:
                    self.viewScene.newPolygon.points.pop(
                        0)  # remove first duplicate point
                self.viewScene.newPolygon = None
                self.viewScene.editModes.setMode("selectItem")
            elif self.viewScene.editModes.addInstance:
                self.viewScene.newInstance = None
                self.viewScene.layoutInstanceTuple = None
                self.viewScene.editModes.setMode("selectItem")
            elif self.viewScene.editModes.addLabel:
                self.viewScene.newLabel = None
                self.viewScene.newLabelTuple = None
                self.viewScene.editModes.setMode("selectItem")
            elif self.viewScene.editModes.cutShape:
                self.viewScene.finishCutLine()
                self.viewScene.editModes.setMode("selectItem")
            elif self.viewScene.editModes.addVia:
                self.viewScene.arrayViaTuple = None
                self.viewScene.arrayVia = None
                self.viewScene.editModes.setMode("selectItem")
        super().keyPressEvent(event)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


from PySide6.QtCore import (
    Qt, Signal,
)
from PySide6.QtWidgets import QLineEdit, QLabel, QWidget


class shortLineEdit(QLineEdit):
    def __init__(self, value: str = ""):
        super().__init__(None)
        self.setMaximumWidth(90)
        if isinstance(value, str):
            self.setText(value)
        else:
            self.setText(str(value))


class boldLabel(QLabel):
    def __init__(self, text: str, parent: QWidget = None):
        super().__init__(text, parent)
        self.setTextFormat(Qt.RichText)
        self.setText("<b>" + text + "</b>")


class longLineEdit(QLineEdit):
    cursorPlaced = Signal()

    def __init__(self, value: str = ""):
        super().__init__(None)
        self.setMaximumWidth(500)
        self.setMinimumWidth(200)
        if isinstance(value, str):
            self.setText(value)
        else:
            self.setText(str(value))

    def focusInEvent(self, event):
        super().focusInEvent(event)
        self.cursorPlaced.emit()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
import pathlib
import re
from pathlib import Path

from PySide6.QtCore import (
    QPoint,
    QRect,
)
from PySide6.QtWidgets import (
    QMainWindow,
)

import revedaEditor.backend.libBackEnd as scb
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.backend.libraryModelView as lmview
import revedaEditor.common.labels as lbl
import revedaEditor.common.shapes as shp
import revedaEditor.fileio.symbolEncoder as symenc
import revedaEditor.gui.symbolEditor as symed
from revedaEditor.backend.pdkLoader import importPDKModule

cb = importPDKModule("callbacks")


class importXschemSym:
    """
    Imports a xschem sym file
    """

    def __init__(
            self,
            parent: QMainWindow,
            filePathObj: Path,
            libraryView: lmview.BaseDesignLibrariesView,
            libraryName: str,
    ):
        self.parent = parent
        self.filePathObj = filePathObj
        self.libraryView = libraryView
        self.libraryName = libraryName
        self._scaleFactor = 4.0
        self._labelHeight = 16
        self._labelXOffset = 40
        self._labelYOffset = 8
        self._labelList = []
        self._pins = []
        self._expressionDict = dict()

        self.cellName = self.filePathObj.stem
        libItem = libm.getLibItem(self.libraryView.libraryModel, self.libraryName)

        cellItem = scb.createCell(self.parent, libItem, self.cellName)
        symbolViewItem = scb.createCellView(self.parent, "symbol", cellItem)
        self.symbolWindow = symed.symbolEditor(
            symbolViewItem, self.parent.libraryDict, self.libraryView
        )
        self.symbolScene = self.symbolWindow.centralW.scene

        self.clbPathObj = pathlib.Path(cb.__file__)
        with self.clbPathObj.open("a") as clbFile:
            clbFile.write("\n\n")
            clbFile.write(f"class {self.cellName}(baseInst):\n")
            clbFile.write("    def __init__(self, labels_dict:dict):\n")
            clbFile.write("        super().__init__(labels_dict)\n")

    def _createScaledPoint(self, x_token, y_token):
        """Helper to create scaled QPoint from string tokens"""
        return QPoint(
            int(self._scaleFactor * float(x_token)), int(self._scaleFactor *
                                                         float(y_token))
        )

    def importSymFile(self):
        # Read file once and process both line-by-line and pattern matching
        with self.filePathObj.open("r") as file:
            fileContent = file.read()

        # Process lines
        lines = fileContent.splitlines()
        processed_lines = []

        for line in lines:
            if not line or len(line) < 1 or line.startswith("*"):
                continue
            if line.startswith("+") and processed_lines:
                processed_lines[-1] += line[1:]  # Append without the '+'
            else:
                if line[0] in "LBPTVGKSE" and line[1] == " ":
                    processed_lines.append(line)
                elif processed_lines:
                    processed_lines[-1] += " " + line

        for line in processed_lines:
            lineTokens = line.split()
            line_type = line[0]
            #
            if line_type == "L" and len(lineTokens) > 4:
                self.symbolScene.lineDraw(
                    self._createScaledPoint(lineTokens[2], lineTokens[3]),
                    self._createScaledPoint(lineTokens[4], lineTokens[5]),
                )

            elif line_type == "B" and len(lineTokens) > 4:
                properties = self.parseLineLine(line)
                point1 = self._createScaledPoint(lineTokens[2], lineTokens[3])
                point2 = self._createScaledPoint(lineTokens[4], lineTokens[5])

                if "name" in properties:
                    pin = shp.symbolPin(
                        QPoint(0, 0),
                        properties.get("name", ""),
                        properties.get("dir", "input").capitalize(),
                        shp.symbolPin.pinTypes[0],
                    )
                    pin.rect = QRect(point1, point2)
                    pin.start = pin.rect.center()
                    self.symbolScene.addItem(pin)
                    self._pins.append(pin)
                else:
                    self.symbolScene.rectDraw(point1, point2)

            elif line_type == "P" and len(lineTokens) > 4:
                numberPoints = int(float(lineTokens[2]))
                points = [
                    self._createScaledPoint(
                        lineTokens[2 * i + 3], lineTokens[2 * i + 4]
                    )
                    for i in range(numberPoints)
                ]
                self.symbolScene.addItem(shp.symbolPolygon(points))

            elif line_type == "K":
                self._expressionDict = self.processFormatString(line)

                if self._expressionDict.get("format"):
                    netlistLine = (
                        self._expressionDict["format"]
                        .replace("@name", "@instName")
                        .replace("@model", "%modelName")
                        .replace("@spiceprefix", "%spiceprefix")
                        .replace("@pinlist", "%pinOrder")
                    )

                    self.symbolScene.attributeList.append(
                        symenc.symbolAttribute("SpiceNetlistLine", netlistLine)
                    )

                templateString = self._expressionDict.get("template")
                textLocation = [self._labelXOffset, self._labelYOffset]
                if templateString:
                    pairs = re.findall(r"(\w+)=([^\s]+)", templateString)
                    templateDict = dict((k, v.strip()) for k, v in pairs)
                    if templateDict:
                        if templateDict.get("name"):
                            templateDict.pop("name")  # we don't use this
                        if templateDict.get("model"):
                            self.symbolScene.attributeList.append(
                                symenc.symbolAttribute(
                                    "modelName", templateDict.pop("model")
                                )
                            )
                        if templateDict.get("spiceprefix"):
                            self.symbolScene.attributeList.append(
                                symenc.symbolAttribute(
                                    "spiceprefix", templateDict.pop("spiceprefix")
                                )
                            )

                    # Add instance name label
                    self._createNLPLabel(textLocation, "[@instName]")

        # Add pin order attribute
        pinNames = [pin.pinName for pin in self._pins]
        self.symbolScene.attributeList.append(
            symenc.symbolAttribute("pinOrder", ", ".join(pinNames))
        )

        self.symbolWindow.checkSaveCell()

    @property
    def scaleFactor(self) -> float:
        return self._scaleFactor

    @scaleFactor.setter
    def scaleFactor(self, value: float):
        self._scaleFactor = value

    @staticmethod
    def parseLineLine(line: str):
        start = line.find("{")
        end = line.find("}", start)
        if start == -1 or end == -1:
            return {}
        return dict(
            pair.split("=", 1) for pair in line[start + 1: end].split() if "=" in pair
        )

    def parseTextLine(self, line: str):
        text = line.split("{")[1].split("}")[0]
        textPropertiesStr = line.split("{")[2].split("}")[0]
        pairs = textPropertiesStr.split()
        textProperties = {}
        for pair in pairs:
            key, value = pair.split("=")
            key = key.strip()
            value = value.strip()
            textProperties[key] = value
        restList = line.split("{")[1].split("}")[1].split()

        textLocation = [
            self._scaleFactor * float(restList[0]),
            self._scaleFactor * float(restList[1]),
        ]
        rotationAngle = float(restList[2])
        return text, textProperties, textLocation, rotationAngle

    @staticmethod
    def processFormatString(inputText: str) -> dict:
        # this is an adhoc list
        keys = {
            "format=": "format",
            "lvs_format=": "lvs",
            "template=": "template",
            "drc=": "drc",
        }
        text = inputText.replace('\\\\"', "").replace("'", "")

        result = {}
        for key_str, key in keys.items():
            pos = text.find(key_str)
            if pos != -1:
                start = pos + len(key_str)
                if start < len(text) and text[start] == '"':
                    start += 1
                    end = text.find('"', start)
                    if end != -1:
                        result[key] = text[start:end]
                else:
                    end = start
                    while end < len(text) and text[end] not in " \n":
                        end += 1
                    result[key] = text[start:end]

        return result

    def _createNLPLabel(self, location: list[int], text: str, add_to_scene=True):
        '''
        Helper method to create and configure NLPLabels with common settings.
        '''
        label = lbl.symbolLabel(QPoint(location[0], location[1]), text,
                                lbl.symbolLabel.labelTypes[1], self._labelHeight,
                                lbl.symbolLabel.labelAlignments[0],
                                lbl.symbolLabel.labelOrients[0],
                                lbl.symbolLabel.labelUses[0])
        label.labelDefs()
        label.labelVisible = True
        label.setOpacity(1)
        if add_to_scene:
            self.symbolScene.addItem(label)
        return label
